# -*- coding: utf-8 -*-
"""Tests for CommandChainDispatcher."""


# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

import pytest
from IPython.core.error import TryNext
from IPython.core.hooks import CommandChainDispatcher

# -----------------------------------------------------------------------------
# Local utilities
# -----------------------------------------------------------------------------


# Define two classes, one which succeeds and one which raises TryNext. Each
# sets the attribute `called` to True when it is called.
class Okay(object):
    def __init__(self, message):
        self.message = message
        self.called = False

    def __call__(self):
        self.called = True
        return self.message


class Fail(object):
    def __init__(self, message):
        self.message = message
        self.called = False

    def __call__(self):
        self.called = True
        raise TryNext(self.message)


# -----------------------------------------------------------------------------
# Test functions
# -----------------------------------------------------------------------------


def test_command_chain_dispatcher_ff():
    """Test two failing hooks"""
    fail1 = Fail("fail1")
    fail2 = Fail("fail2")
    dp = CommandChainDispatcher([(0, fail1), (10, fail2)])

    with pytest.raises(TryNext) as e:
        dp()
    assert str(e.value) == "fail2"

    assert fail1.called is True
    assert fail2.called is True


def test_command_chain_dispatcher_fofo():
    """Test a mixture of failing and succeeding hooks."""
    fail1 = Fail("fail1")
    fail2 = Fail("fail2")
    okay1 = Okay("okay1")
    okay2 = Okay("okay2")

    dp = CommandChainDispatcher(
        [
            (0, fail1),
            # (5, okay1), # add this later
            (10, fail2),
            (15, okay2),
        ]
    )
    dp.add(okay1, 5)

    assert dp() == "okay1"

    assert fail1.called is True
    assert okay1.called is True
    assert fail2.called is False
    assert okay2.called is False


def test_command_chain_dispatcher_eq_priority():
    okay1 = Okay("okay1")
    okay2 = Okay("okay2")
    dp = CommandChainDispatcher([(1, okay1)])
    dp.add(okay2, 1)



"""Test embedding of IPython"""

# -----------------------------------------------------------------------------
#  Copyright (C) 2013 The IPython Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file COPYING, distributed as part of this software.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

import os
import subprocess
import sys

from IPython.utils.tempdir import NamedFileInTemporaryDirectory
from IPython.testing.decorators import skip_win32
from IPython.testing import IPYTHON_TESTING_TIMEOUT_SCALE

# -----------------------------------------------------------------------------
# Tests
# -----------------------------------------------------------------------------


_sample_embed = """
import IPython

a = 3
b = 14
print(a, '.', b)

IPython.embed()

print('bye!')
"""

_exit = "exit\r"


def test_ipython_embed():
    """test that `IPython.embed()` works"""
    with NamedFileInTemporaryDirectory("file_with_embed.py", "w") as f:
        f.write(_sample_embed)
        f.flush()
        f.close()  # otherwise msft won't be able to read the file

        # run `python file_with_embed.py`
        cmd = [sys.executable, f.name]
        env = os.environ.copy()
        env["IPY_TEST_SIMPLE_PROMPT"] = "1"

        p = subprocess.Popen(
            cmd,
            env=env,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            encoding="UTF-8",
        )
        std, err = p.communicate(_exit)
        assert isinstance(std, str), (std, err, p.returncode)

        assert p.returncode == 0, (p.returncode, std)
        assert "3 . 14" in std
        if os.name != "nt":
            # TODO: Fix up our different stdout references, see issue gh-14
            assert "IPython" in std
        assert "bye!" in std


@skip_win32
def test_nest_embed():
    """test that `IPython.embed()` is nestable"""
    import pexpect

    ipy_prompt = r"]:"  # ansi color codes give problems matching beyond this
    env = os.environ.copy()
    env["IPY_TEST_SIMPLE_PROMPT"] = "1"

    child = pexpect.spawn(
        sys.executable, ["-m", "IPython", "--colors=nocolor"], env=env
    )
    child.timeout = 15 * IPYTHON_TESTING_TIMEOUT_SCALE
    child.expect(ipy_prompt)
    child.timeout = 5 * IPYTHON_TESTING_TIMEOUT_SCALE
    child.sendline("import IPython")
    child.sendline("from IPython.core.history import HistoryManager")
    child.sendline("HistoryManager._max_inst = 3")
    child.expect(ipy_prompt)
    child.sendline("ip0 = get_ipython()")
    # enter first nested embed
    child.sendline("IPython.embed()")
    # skip the banner until we get to a prompt
    try:
        prompted = -1
        while prompted != 0:
            prompted = child.expect([ipy_prompt, "\r\n"])
    except pexpect.TIMEOUT as e:
        print(e)
        # child.interact()
    child.sendline("embed1 = get_ipython()")
    child.expect(ipy_prompt)
    child.sendline("print('true' if embed1 is not ip0 else 'false')")
    assert child.expect(["true\r\n", "false\r\n"]) == 0
    child.expect(ipy_prompt)
    child.sendline("print('true' if IPython.get_ipython() is embed1 else 'false')")
    assert child.expect(["true\r\n", "false\r\n"]) == 0
    child.expect(ipy_prompt)
    # enter second nested embed
    child.sendline("IPython.embed()")
    # skip the banner until we get to a prompt
    try:
        prompted = -1
        while prompted != 0:
            prompted = child.expect([ipy_prompt, "\r\n"])
    except pexpect.TIMEOUT as e:
        print(e)
        # child.interact()
    child.sendline("embed2 = get_ipython()")
    child.expect(ipy_prompt)
    child.sendline("print('true' if embed2 is not embed1 else 'false')")
    assert child.expect(["true\r\n", "false\r\n"]) == 0
    child.expect(ipy_prompt)
    child.sendline("print('true' if embed2 is IPython.get_ipython() else 'false')")
    assert child.expect(["true\r\n", "false\r\n"]) == 0
    child.expect(ipy_prompt)
    child.sendline("exit")
    # back at first embed
    child.expect(ipy_prompt)
    child.sendline("print('true' if get_ipython() is embed1 else 'false')")
    assert child.expect(["true\r\n", "false\r\n"]) == 0
    child.expect(ipy_prompt)
    child.sendline("print('true' if IPython.get_ipython() is embed1 else 'false')")
    assert child.expect(["true\r\n", "false\r\n"]) == 0
    child.expect(ipy_prompt)
    child.sendline("exit")
    # back at launching scope
    child.expect(ipy_prompt)
    child.sendline("print('true' if get_ipython() is ip0 else 'false')")
    assert child.expect(["true\r\n", "false\r\n"]) == 0
    child.expect(ipy_prompt)
    child.sendline("print('true' if IPython.get_ipython() is ip0 else 'false')")
    assert child.expect(["true\r\n", "false\r\n"]) == 0
    child.expect(ipy_prompt)
    child.sendline("exit")
    child.close()



"""Embed IPython using the simple embed function rather than the class API."""

from IPython import embed

a = 10
b = 20

embed(header="First time", banner1="")

c = 30
d = 40

try:
    raise Exception("adsfasdf")
except:
    embed(header="The second time")



"""Some tests for the wildcard utilities."""

# -----------------------------------------------------------------------------
# Library imports
# -----------------------------------------------------------------------------
# Stdlib
import unittest

# Our own
from IPython.utils import wildcard

# -----------------------------------------------------------------------------
# Globals for test
# -----------------------------------------------------------------------------


class obj_t(object):
    pass


root = obj_t()
l = ["arna", "abel", "ABEL", "active", "bob", "bark", "abbot"]
q = ["kate", "loop", "arne", "vito", "lucifer", "koppel"]
for x in l:
    o = obj_t()
    setattr(root, x, o)
    for y in q:
        p = obj_t()
        setattr(o, y, p)
root._apan = obj_t()
root._apan.a = 10
root._apan._a = 20
root._apan.__a = 20
root.__anka = obj_t()
root.__anka.a = 10
root.__anka._a = 20
root.__anka.__a = 20

root._APAN = obj_t()
root._APAN.a = 10
root._APAN._a = 20
root._APAN.__a = 20
root.__ANKA = obj_t()
root.__ANKA.a = 10
root.__ANKA._a = 20
root.__ANKA.__a = 20

# -----------------------------------------------------------------------------
# Test cases
# -----------------------------------------------------------------------------


class Tests(unittest.TestCase):
    def test_case(self):
        ns = root.__dict__
        tests = [
            (
                "a*",
                [
                    "abbot",
                    "abel",
                    "active",
                    "arna",
                ],
            ),
            (
                "?b*.?o*",
                [
                    "abbot.koppel",
                    "abbot.loop",
                    "abel.koppel",
                    "abel.loop",
                ],
            ),
            ("_a*", []),
            (
                "_*anka",
                [
                    "__anka",
                ],
            ),
            (
                "_*a*",
                [
                    "__anka",
                ],
            ),
        ]
        for pat, res in tests:
            res.sort()
            a = sorted(
                wildcard.list_namespace(
                    ns, "all", pat, ignore_case=False, show_all=False
                ).keys()
            )
            self.assertEqual(a, res)

    def test_case_showall(self):
        ns = root.__dict__
        tests = [
            (
                "a*",
                [
                    "abbot",
                    "abel",
                    "active",
                    "arna",
                ],
            ),
            (
                "?b*.?o*",
                [
                    "abbot.koppel",
                    "abbot.loop",
                    "abel.koppel",
                    "abel.loop",
                ],
            ),
            ("_a*", ["_apan"]),
            (
                "_*anka",
                [
                    "__anka",
                ],
            ),
            (
                "_*a*",
                [
                    "__anka",
                    "_apan",
                ],
            ),
        ]
        for pat, res in tests:
            res.sort()
            a = sorted(
                wildcard.list_namespace(
                    ns, "all", pat, ignore_case=False, show_all=True
                ).keys()
            )
            self.assertEqual(a, res)

    def test_nocase(self):
        ns = root.__dict__
        tests = [
            (
                "a*",
                [
                    "abbot",
                    "abel",
                    "ABEL",
                    "active",
                    "arna",
                ],
            ),
            (
                "?b*.?o*",
                [
                    "abbot.koppel",
                    "abbot.loop",
                    "abel.koppel",
                    "abel.loop",
                    "ABEL.koppel",
                    "ABEL.loop",
                ],
            ),
            ("_a*", []),
            (
                "_*anka",
                [
                    "__anka",
                    "__ANKA",
                ],
            ),
            (
                "_*a*",
                [
                    "__anka",
                    "__ANKA",
                ],
            ),
        ]
        for pat, res in tests:
            res.sort()
            a = sorted(
                wildcard.list_namespace(
                    ns, "all", pat, ignore_case=True, show_all=False
                ).keys()
            )
            self.assertEqual(a, res)

    def test_nocase_showall(self):
        ns = root.__dict__
        tests = [
            (
                "a*",
                [
                    "abbot",
                    "abel",
                    "ABEL",
                    "active",
                    "arna",
                ],
            ),
            (
                "?b*.?o*",
                [
                    "abbot.koppel",
                    "abbot.loop",
                    "abel.koppel",
                    "abel.loop",
                    "ABEL.koppel",
                    "ABEL.loop",
                ],
            ),
            ("_a*", ["_apan", "_APAN"]),
            (
                "_*anka",
                [
                    "__anka",
                    "__ANKA",
                ],
            ),
            ("_*a*", ["__anka", "__ANKA", "_apan", "_APAN"]),
        ]
        for pat, res in tests:
            res.sort()
            a = sorted(
                wildcard.list_namespace(
                    ns, "all", pat, ignore_case=True, show_all=True
                ).keys()
            )
            a.sort()
            self.assertEqual(a, res)

    def test_dict_attributes(self):
        """Dictionaries should be indexed by attributes, not by keys. This was
        causing Github issue 129."""
        ns = {"az": {"king": 55}, "pq": {1: 0}}
        tests = [("a*", ["az"]), ("az.k*", ["az.keys"]), ("pq.k*", ["pq.keys"])]
        for pat, res in tests:
            res.sort()
            a = sorted(
                wildcard.list_namespace(
                    ns, "all", pat, ignore_case=False, show_all=True
                ).keys()
            )
            self.assertEqual(a, res)

    def test_dict_dir(self):
        class A(object):
            def __init__(self):
                self.a = 1
                self.b = 2

            def __getattribute__(self, name):
                if name == "a":
                    raise AttributeError
                return object.__getattribute__(self, name)

        a = A()
        adict = wildcard.dict_dir(a)
        assert "a" not in adict  # change to assertNotIn method in >= 2.7
        self.assertEqual(adict["b"], 2)



# -*- coding: utf-8 -*-
"""Tests for the TerminalInteractiveShell and related pieces."""
# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import sys
import unittest
import os

from prompt_toolkit.auto_suggest import AutoSuggestFromHistory


from IPython.testing import tools as tt

from IPython.terminal.ptutils import _elide, _adjust_completion_text_based_on_context
from IPython.terminal.shortcuts.auto_suggest import NavigableAutoSuggestFromHistory


class TestAutoSuggest(unittest.TestCase):
    def test_changing_provider(self):
        ip = get_ipython()
        ip.autosuggestions_provider = None
        self.assertEqual(ip.auto_suggest, None)
        ip.autosuggestions_provider = "AutoSuggestFromHistory"
        self.assertIsInstance(ip.auto_suggest, AutoSuggestFromHistory)
        ip.autosuggestions_provider = "NavigableAutoSuggestFromHistory"
        self.assertIsInstance(ip.auto_suggest, NavigableAutoSuggestFromHistory)


def test_elide():
    _elide("concatenate((a1, a2, ...), axis", "", min_elide=30)  # do not raise
    _elide("concatenate((a1, a2, ..), . axis", "", min_elide=30)  # do not raise
    assert (
        _elide("aaaa.bbbb.ccccc.dddddd.eeeee.fffff.gggggg.hhhhhh", "", min_elide=30)
        == "aaaa.b…g.hhhhhh"
    )

    test_string = os.sep.join(["", 10 * "a", 10 * "b", 10 * "c", ""])
    expect_string = os.sep + "a" + "\N{HORIZONTAL ELLIPSIS}" + "b" + os.sep + 10 * "c"
    assert _elide(test_string, "", min_elide=30) == expect_string


def test_elide_typed_normal():
    assert (
        _elide(
            "the quick brown fox jumped over the lazy dog",
            "the quick brown fox",
            min_elide=10,
        )
        == "the…fox jumped over the lazy dog"
    )


def test_elide_typed_short_match():
    """
    if the match is too short we don't elide.
    avoid the "the...the"
    """
    assert (
        _elide("the quick brown fox jumped over the lazy dog", "the", min_elide=10)
        == "the quick brown fox jumped over the lazy dog"
    )


def test_elide_typed_no_match():
    """
    if the match is too short we don't elide.
    avoid the "the...the"
    """
    # here we typed red instead of brown
    assert (
        _elide(
            "the quick brown fox jumped over the lazy dog",
            "the quick red fox",
            min_elide=10,
        )
        == "the quick brown fox jumped over the lazy dog"
    )


class TestContextAwareCompletion(unittest.TestCase):
    def test_adjust_completion_text_based_on_context(self):
        # Adjusted case
        self.assertEqual(
            _adjust_completion_text_based_on_context("arg1=", "func1(a=)", 7), "arg1"
        )

        # Untouched cases
        self.assertEqual(
            _adjust_completion_text_based_on_context("arg1=", "func1(a)", 7), "arg1="
        )
        self.assertEqual(
            _adjust_completion_text_based_on_context("arg1=", "func1(a", 7), "arg1="
        )
        self.assertEqual(
            _adjust_completion_text_based_on_context("%magic", "func1(a=)", 7), "%magic"
        )
        self.assertEqual(
            _adjust_completion_text_based_on_context("func2", "func1(a=)", 7), "func2"
        )


# Decorator for interaction loop tests -----------------------------------------


class mock_input_helper(object):
    """Machinery for tests of the main interact loop.

    Used by the mock_input decorator.
    """

    def __init__(self, testgen):
        self.testgen = testgen
        self.exception = None
        self.ip = get_ipython()

    def __enter__(self):
        self.orig_prompt_for_code = self.ip.prompt_for_code
        self.ip.prompt_for_code = self.fake_input
        return self

    def __exit__(self, etype, value, tb):
        self.ip.prompt_for_code = self.orig_prompt_for_code

    def fake_input(self):
        try:
            return next(self.testgen)
        except StopIteration:
            self.ip.keep_running = False
            return ""
        except:
            self.exception = sys.exc_info()
            self.ip.keep_running = False
            return ""


def mock_input(testfunc):
    """Decorator for tests of the main interact loop.

    Write the test as a generator, yield-ing the input strings, which IPython
    will see as if they were typed in at the prompt.
    """

    def test_method(self):
        testgen = testfunc(self)
        with mock_input_helper(testgen) as mih:
            mih.ip.interact()

        if mih.exception is not None:
            # Re-raise captured exception
            etype, value, tb = mih.exception
            import traceback

            traceback.print_tb(tb, file=sys.stdout)
            del tb  # Avoid reference loop
            raise value

    return test_method


# Test classes -----------------------------------------------------------------


class InteractiveShellTestCase(unittest.TestCase):
    def rl_hist_entries(self, rl, n):
        """Get last n readline history entries as a list"""
        return [
            rl.get_history_item(rl.get_current_history_length() - x)
            for x in range(n - 1, -1, -1)
        ]

    @mock_input
    def test_inputtransformer_syntaxerror(self):
        ip = get_ipython()
        ip.input_transformers_post.append(syntax_error_transformer)

        try:
            # raise Exception
            with tt.AssertPrints("4", suppress=False):
                yield "print(2*2)"

            with tt.AssertPrints("SyntaxError: input contains", suppress=False):
                yield "print(2345) # syntaxerror"

            with tt.AssertPrints("16", suppress=False):
                yield "print(4*4)"

        finally:
            ip.input_transformers_post.remove(syntax_error_transformer)

    def test_repl_not_plain_text(self):
        ip = get_ipython()
        formatter = ip.display_formatter
        assert formatter.active_types == ["text/plain"]

        # terminal may have arbitrary mimetype handler to open external viewer
        # or inline images.
        assert formatter.ipython_display_formatter.enabled

        class Test(object):
            def __repr__(self):
                return "<Test %i>" % id(self)

            def _repr_html_(self):
                return "<html>"

        # verify that HTML repr isn't computed
        obj = Test()
        data, _ = formatter.format(obj)
        self.assertEqual(data, {"text/plain": repr(obj)})

        class Test2(Test):
            def _ipython_display_(self):
                from IPython.display import display, HTML

                display(HTML("<custom>"))

        # verify that mimehandlers are called
        called = False

        def handler(data, metadata):
            print("Handler called")
            nonlocal called
            called = True

        ip.display_formatter.active_types.append("text/html")
        ip.display_formatter.formatters["text/html"].enabled = True
        ip.mime_renderers["text/html"] = handler
        try:
            obj = Test()
            display(obj)
        finally:
            ip.display_formatter.formatters["text/html"].enabled = False
            del ip.mime_renderers["text/html"]

        assert called == True


def syntax_error_transformer(lines):
    """Transformer that throws SyntaxError if 'syntaxerror' is in the code."""
    for line in lines:
        pos = line.find("syntaxerror")
        if pos >= 0:
            e = SyntaxError('input contains "syntaxerror"')
            e.text = line
            e.offset = pos + 1
            raise e
    return lines


class TerminalMagicsTestCase(unittest.TestCase):
    def test_paste_magics_blankline(self):
        """Test that code with a blank line doesn't get split (gh-3246)."""
        ip = get_ipython()
        s = "def pasted_func(a):\n" "    b = a+1\n" "\n" "    return b"

        tm = ip.magics_manager.registry["TerminalMagics"]
        tm.store_or_execute(s, name=None)

        self.assertEqual(ip.user_ns["pasted_func"](54), 55)



# coding: utf-8

# This script autogenerates `IPython.core.latex_symbols.py`, which contains a
# single dict , named `latex_symbols`. The keys in this dict are latex symbols,
# such as `\\alpha` and the values in the dict are the unicode equivalents for
# those. Most importantly, only unicode symbols that are valid identifiers in
# Python 3 are included.

#
# The original mapping of latex symbols to unicode comes from the `latex_symbols.jl` files from Julia.

from pathlib import Path

# Import the Julia LaTeX symbols
print("Importing latex_symbols.jl from Julia...")
import requests

url = "https://raw.githubusercontent.com/JuliaLang/julia/master/stdlib/REPL/src/latex_symbols.jl"
r = requests.get(url)


# Build a list of key, value pairs
print("Building a list of (latex, unicode) key-value pairs...")
lines = r.text.splitlines()

prefixes_line = lines.index('# "font" prefixes')
symbols_line = lines.index("# manual additions:")

prefix_dict = {}
for l in lines[prefixes_line + 1 : symbols_line]:
    p = l.split()
    if not p or p[1] == "latex_symbols":
        continue
    prefix_dict[p[1]] = p[3]

idents = []
for l in lines[symbols_line:]:
    if "=>" not in l:
        continue  # if it's not a def, skip
    if "#" in l:
        l = l[: l.index("#")]  # get rid of eol comments
    x, y = l.strip().split("=>")
    if "*" in x:  # if a prefix is present substitute it with its value
        p, x = x.split("*")
        x = prefix_dict[p][:-1] + x[1:]
    try:
        x = x.split('"')[1]
        y = y.split('"')[1]  # get the values in quotes
    except IndexError as e:
        continue
        raise ValueError((x, y)) from e
    if not x.startswith("\\\\"):
        # reverse mapping
        continue
    idents.append((x, y))

# Filter out non-valid identifiers
print("Filtering out characters that are not valid Python 3 identifiers")


def test_ident(i):
    """Is the unicode string valid in a Python 3 identifier."""
    # Some characters are not valid at the start of a name, but we still want to
    # include them. So prefix with 'a', which is valid at the start.
    return ("a" + i).isidentifier()


assert test_ident("α")
assert not test_ident("‴")

valid_idents = [line for line in idents if test_ident(line[1])]

# Write the `latex_symbols.py` module in the cwd

s = f"""# encoding: utf-8

# DO NOT EDIT THIS FILE BY HAND.

# To update this file, run the script /tools/gen_latex_symbols.py using Python 3

# This file is autogenerated from the file:
# {url}
# This original list is filtered to remove any unicode characters that are not valid
# Python identifiers.

latex_symbols = {{
"""
for line in valid_idents:
    s += '    "%s": "%s",\n' % (line[0], line[1])
s += "}\n"

s += """

reverse_latex_symbol = {v: k for k, v in latex_symbols.items()}
"""

HERE = Path(__file__).parent

fn = HERE / ".." / "IPython" / "core" / "latex_symbols.py"
print("Writing the file: %s" % str(fn))
fn.write_text(s, encoding="utf-8")



"""Test embedding of IPython"""

# -----------------------------------------------------------------------------
#  Copyright (C) 2013 The IPython Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file COPYING, distributed as part of this software.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

import os
import sys
from IPython.testing.decorators import skip_win32
from IPython.testing import IPYTHON_TESTING_TIMEOUT_SCALE

# -----------------------------------------------------------------------------
# Tests
# -----------------------------------------------------------------------------


@skip_win32
def test_debug_magic_passes_through_generators():
    """
    This test that we can correctly pass through frames of a generator post-mortem.
    """
    import pexpect
    import re

    in_prompt = re.compile(rb"In ?\[\d+\]:")
    ipdb_prompt = "ipdb>"
    env = os.environ.copy()
    child = pexpect.spawn(
        sys.executable,
        ["-m", "IPython", "--colors=nocolor", "--simple-prompt"],
        env=env,
    )
    child.timeout = 15 * IPYTHON_TESTING_TIMEOUT_SCALE

    child.expect(in_prompt)

    child.timeout = 2 * IPYTHON_TESTING_TIMEOUT_SCALE

    child.sendline("def f(x):")
    child.sendline("    raise Exception")
    child.sendline("")

    child.expect(in_prompt)
    child.sendline("gen = (f(x) for x in [0])")
    child.sendline("")

    child.expect(in_prompt)
    child.sendline("for x in gen:")
    child.sendline("    pass")
    child.sendline("")

    child.timeout = 10 * IPYTHON_TESTING_TIMEOUT_SCALE

    child.expect("Exception:")

    child.expect(in_prompt)
    child.sendline(r"%debug")
    child.expect("----> 2     raise Exception")

    child.expect(ipdb_prompt)
    child.sendline("u")
    child.expect_exact(r"----> 1 gen = (f(x) for x in [0])")

    child.expect(ipdb_prompt)
    child.sendline("u")
    child.expect_exact("----> 1 for x in gen:")

    child.expect(ipdb_prompt)
    child.sendline("u")
    child.expect_exact(
        "*** all frames above skipped (hidden frames and ignored modules). Use `skip_hidden False` for hidden frames or unignore_module for ignored modules."
    )

    child.expect(ipdb_prompt)
    child.sendline("exit")

    child.expect(in_prompt)
    child.sendline("exit")

    child.close()



"""Test installing editor hooks"""

import sys
from unittest import mock

from IPython import get_ipython
from IPython.lib import editorhooks


def test_install_editor():
    called = []

    def fake_popen(*args, **kwargs):
        called.append(
            {
                "args": args,
                "kwargs": kwargs,
            }
        )
        return mock.MagicMock(**{"wait.return_value": 0})

    editorhooks.install_editor("foo -l {line} -f {filename}", wait=False)

    with mock.patch("subprocess.Popen", fake_popen):
        get_ipython().hooks.editor("the file", 64)

    assert len(called) == 1
    args = called[0]["args"]
    kwargs = called[0]["kwargs"]

    assert kwargs == {"shell": True}

    if sys.platform.startswith("win"):
        expected = ["foo", "-l", "64", "-f", "the file"]
    else:
        expected = "foo -l 64 -f 'the file'"
    cmd = args[0]
    assert cmd == expected



# -*- coding: utf-8 -*-
"""Tests for completerlib."""

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

import os
import shutil
import sys
import tempfile
import unittest
from os.path import join

from tempfile import TemporaryDirectory

from IPython.core.completerlib import magic_run_completer, module_completion, try_import
from IPython.testing.decorators import onlyif_unicode_paths


class MockEvent(object):
    def __init__(self, line):
        self.line = line


# -----------------------------------------------------------------------------
# Test functions begin
# -----------------------------------------------------------------------------
class Test_magic_run_completer(unittest.TestCase):
    files = ["aao.py", "a.py", "b.py", "aao.txt"]
    dirs = ["adir/", "bdir/"]

    def setUp(self):
        self.BASETESTDIR = tempfile.mkdtemp()
        for fil in self.files:
            with open(join(self.BASETESTDIR, fil), "w", encoding="utf-8") as sfile:
                sfile.write("pass\n")
        for d in self.dirs:
            os.mkdir(join(self.BASETESTDIR, d))

        self.oldpath = os.getcwd()
        os.chdir(self.BASETESTDIR)

    def tearDown(self):
        os.chdir(self.oldpath)
        shutil.rmtree(self.BASETESTDIR)

    def test_1(self):
        """Test magic_run_completer, should match two alternatives"""
        event = MockEvent("%run a")
        mockself = None
        match = set(magic_run_completer(mockself, event))
        self.assertEqual(match, {"a.py", "aao.py", "adir/"})

    def test_2(self):
        """Test magic_run_completer, should match one alternative"""
        event = MockEvent("%run aa")
        mockself = None
        match = set(magic_run_completer(mockself, event))
        self.assertEqual(match, {"aao.py"})

    def test_3(self):
        """Test magic_run_completer with unterminated " """
        event = MockEvent('%run "a')
        mockself = None
        match = set(magic_run_completer(mockself, event))
        self.assertEqual(match, {"a.py", "aao.py", "adir/"})

    def test_completion_more_args(self):
        event = MockEvent("%run a.py ")
        match = set(magic_run_completer(None, event))
        self.assertEqual(match, set(self.files + self.dirs))

    def test_completion_in_dir(self):
        # Github issue #3459
        event = MockEvent("%run a.py {}".format(join(self.BASETESTDIR, "a")))
        print(repr(event.line))
        match = set(magic_run_completer(None, event))
        # We specifically use replace here rather than normpath, because
        # at one point there were duplicates 'adir' and 'adir/', and normpath
        # would hide the failure for that.
        self.assertEqual(
            match,
            {
                join(self.BASETESTDIR, f).replace("\\", "/")
                for f in ("a.py", "aao.py", "aao.txt", "adir/")
            },
        )


class Test_magic_run_completer_nonascii(unittest.TestCase):
    @onlyif_unicode_paths
    def setUp(self):
        self.BASETESTDIR = tempfile.mkdtemp()
        for fil in ["aaø.py", "a.py", "b.py"]:
            with open(join(self.BASETESTDIR, fil), "w", encoding="utf-8") as sfile:
                sfile.write("pass\n")
        self.oldpath = os.getcwd()
        os.chdir(self.BASETESTDIR)

    def tearDown(self):
        os.chdir(self.oldpath)
        shutil.rmtree(self.BASETESTDIR)

    @onlyif_unicode_paths
    def test_1(self):
        """Test magic_run_completer, should match two alternatives"""
        event = MockEvent("%run a")
        mockself = None
        match = set(magic_run_completer(mockself, event))
        self.assertEqual(match, {"a.py", "aaø.py"})

    @onlyif_unicode_paths
    def test_2(self):
        """Test magic_run_completer, should match one alternative"""
        event = MockEvent("%run aa")
        mockself = None
        match = set(magic_run_completer(mockself, event))
        self.assertEqual(match, {"aaø.py"})

    @onlyif_unicode_paths
    def test_3(self):
        """Test magic_run_completer with unterminated " """
        event = MockEvent('%run "a')
        mockself = None
        match = set(magic_run_completer(mockself, event))
        self.assertEqual(match, {"a.py", "aaø.py"})


# module_completer:


def test_import_invalid_module():
    """Testing of issue https://github.com/ipython/ipython/issues/1107"""
    invalid_module_names = {"foo-bar", "foo:bar", "10foo"}
    valid_module_names = {"foobar"}
    with TemporaryDirectory() as tmpdir:
        sys.path.insert(0, tmpdir)
        for name in invalid_module_names | valid_module_names:
            filename = os.path.join(tmpdir, name + ".py")
            open(filename, "w", encoding="utf-8").close()

        s = set(module_completion("import foo"))
        intersection = s.intersection(invalid_module_names)
        assert intersection == set()

        assert valid_module_names.issubset(s), valid_module_names.intersection(s)


def test_bad_module_all():
    """Test module with invalid __all__

    https://github.com/ipython/ipython/issues/9678
    """
    testsdir = os.path.dirname(__file__)
    sys.path.insert(0, testsdir)
    try:
        results = module_completion("from bad_all import ")
        assert "puppies" in results
        for r in results:
            assert isinstance(r, str)

        # bad_all doesn't contain submodules, but this completion
        # should finish without raising an exception:
        results = module_completion("import bad_all.")
        assert results == []
    finally:
        sys.path.remove(testsdir)


def test_module_without_init():
    """
    Test module without __init__.py.

    https://github.com/ipython/ipython/issues/11226
    """
    fake_module_name = "foo_xder_134"
    with TemporaryDirectory() as tmpdir:
        sys.path.insert(0, tmpdir)
        try:
            os.makedirs(os.path.join(tmpdir, fake_module_name))
            s = try_import(mod=fake_module_name)
            assert s == [], f"for module {fake_module_name}"
        finally:
            sys.path.remove(tmpdir)


def test_valid_exported_submodules():
    """
    Test checking exported (__all__) objects are submodules
    """
    results = module_completion("import os.pa")
    # ensure we get a valid submodule:
    assert "os.path" in results
    # ensure we don't get objects that aren't submodules:
    assert "os.pathconf" not in results



"""Tests for tokenutil"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import pytest
import textwrap

from IPython.utils.tokenutil import token_at_cursor, line_at_cursor


def expect_token(expected: str, cell: str, cursor_pos: int | None = None) -> None:
    """
    If cursor_pos is `None`, look for `|` as the cursor position.
    Assert that the token at the cursor position is `expected`.
    """
    if cursor_pos is None:
        assert (
            cursor_count := cell.count("|")
        ) == 1, (
            f"Cursor position not specified and found {cursor_count} instance(s) of '|'"
        )
        cursor_pos = cell.index("|")
        cell = cell.replace("|", "")
    token = token_at_cursor(cell, cursor_pos)
    offset = 0
    for line in cell.splitlines():
        if offset + len(line) >= cursor_pos:
            break
        else:
            offset += len(line) + 1
    column = cursor_pos - offset
    line_with_cursor = "%s|%s" % (line[:column], line[column:])
    assert token == expected, "Expected %r, got %r in: %r (pos %i)" % (
        expected,
        token,
        line_with_cursor,
        cursor_pos,
    )


def test_simple():
    cell = "foo"
    for i in range(len(cell)):
        expect_token("foo", cell, i)


def test_function():
    cell = "foo(a=5, b='10')"
    for i in range(len(cell)):
        expect_token("foo", cell, i)
    expect_token("a", "foo(|a, b)")
    expect_token("foo", "foo(a|, b)")
    expect_token("foo", "foo(a,| b)")
    expect_token("b", "foo(a, |b)")
    expect_token("foo", "foo(a, b|)")


@pytest.mark.parametrize(
    "cell",
    [
        "\n".join(["a = 5", "b = hello('string', there)"]),
        textwrap.dedent(
            '''
            """\n\nxxxxxxxxxx\n\n"""
            5, """
            """\n\nxxxxxxxxxx\n\n"""
            docstring
            multiline token
            """, [
            2, 3, "complicated"]
            b = hello("string", there)
        '''
        ),
    ],
)
def test_multiline(cell):
    expected = "hello"
    start = cell.index(expected) + 1
    for i in range(start, start + len(expected)):
        expect_token(expected, cell, i)
    start = cell.index(expected) + 1
    for i in range(start, start + len(expected)):
        expect_token(expected, cell, i)


def test_nested_call_kwargs():
    cell = "foo(bar(a=5), b=10)"
    start = cell.index("bar")
    last = start + len("bar")
    for i in range(start, last + 1):
        expect_token("bar", cell, i)
    start = cell.index("a=")
    for i in range(start, start + 3):
        expect_token("bar", cell, i)
    expect_token("bar", "foo(bar(a=|5), b=10)")
    expect_token("bar", "foo(bar(a=5|), b=10)")
    for i in range(cell.index(")") + 1, cell.index("b=") + 2):
        expect_token("foo", cell, i)
    expect_token("foo", cell, len(cell) - 1)


def test_nested_call_args():
    expect_token("bar", "foo(bar(x,| ))")
    expect_token("bar", "foo(bar(x, |))")
    expect_token("bar", "foo(ba|r(x, y))")
    expect_token("x", "foo(bar(|x, y))")
    expect_token("bar", "foo(bar(x|, y))")
    expect_token("bar", "foo(bar(x,| y))")
    expect_token("y", "foo(bar(x, |y))")
    expect_token("bar", "foo(bar(x, y|))")
    expect_token("foo", "foo(bar(x, y)|)")
    expect_token("foo", "foo(bar(, y)|)")


def test_outer_name():
    expect_token("x", "|x + foo(a, b) + y + z")
    expect_token("x", "x| + foo(a, b) + y + z")
    expect_token("x", "x |+ foo(a, b) + y + z")
    expect_token("x", "x +| foo(a, b) + y + z")
    expect_token("foo", "x + |foo(a, b) + y + z")
    expect_token("a", "x + foo(|a, b) + y + z")
    expect_token("foo", "x + foo(a|, b) + y + z")
    expect_token("foo", "x + foo(a, b)| + y + z")
    expect_token("foo", "x + foo(a, b) |+ y + z")
    expect_token("foo", "x + foo(a, b) +| y + z")
    expect_token("y", "x + foo(a, b) + |y + z")
    expect_token("y", "x + foo(a, b) + y| + z")
    expect_token("y", "x + foo(a, b) + y |+ z")
    expect_token("y", "x + foo(a, b) + y +| z")
    expect_token("z", "x + foo(a, b) + y + |z")


def test_attrs():
    cell = "a = obj.attr.subattr"
    expected = "obj"
    idx = cell.find("obj") + 1
    for i in range(idx, idx + 3):
        expect_token(expected, cell, i)
    idx = cell.find(".attr") + 2
    expected = "obj.attr"
    for i in range(idx, idx + 4):
        expect_token(expected, cell, i)
    idx = cell.find(".subattr") + 2
    expected = "obj.attr.subattr"
    for i in range(idx, len(cell)):
        expect_token(expected, cell, i)


def test_line_at_cursor():
    cell = ""
    (line, offset) = line_at_cursor(cell, cursor_pos=11)
    assert line == ""
    assert offset == 0

    # The position after a newline should be the start of the following line.
    cell = "One\nTwo\n"
    (line, offset) = line_at_cursor(cell, cursor_pos=4)
    assert line == "Two\n"
    assert offset == 4

    # The end of a cell should be on the last line
    cell = "pri\npri"
    (line, offset) = line_at_cursor(cell, cursor_pos=7)
    assert line == "pri"
    assert offset == 4


@pytest.mark.parametrize(
    "c, token",
    zip(
        list(range(16, 22)) + list(range(22, 28)),
        ["int"] * (22 - 16) + ["map"] * (28 - 22),
    ),
)
def test_multiline_statement(c, token):
    cell = """a = (1,
    3)

int()
map()
"""
    expect_token(token, cell, c)



"""Tests for the line-based transformers in IPython.core.inputtransformer2

Line-based transformers are the simpler ones; token-based transformers are
more complex. See test_inputtransformer2 for tests for token-based transformers.
"""

import pytest

from IPython.core import inputtransformer2 as ipt2

CELL_MAGIC = (
    """\
%%foo arg
body 1
body 2
""",
    """\
get_ipython().run_cell_magic('foo', 'arg', 'body 1\\nbody 2\\n')
""",
)


def test_cell_magic():
    for sample, expected in [CELL_MAGIC]:
        assert ipt2.cell_magic(sample.splitlines(keepends=True)) == expected.splitlines(
            keepends=True
        )


CLASSIC_PROMPT = (
    """\
>>> for a in range(5):
...     print(a)
""",
    """\
for a in range(5):
    print(a)
""",
)

CLASSIC_PROMPT_L2 = (
    """\
for a in range(5):
...     print(a)
...     print(a ** 2)
""",
    """\
for a in range(5):
...     print(a)
...     print(a ** 2)
""",
)

CLASSIC_PROMPT_L3 = (
    """\
>>> \"\"\"
... This code is inside a triple-quoted string.
... >>> for a in range(5):
... ...     print(a)
... \"\"\"
>>> for a in range(5):
...     print(a)
""",
    """\
>>> \"\"\"
... This code is inside a triple-quoted string.
... >>> for a in range(5):
... ...     print(a)
... \"\"\"
for a in range(5):
    print(a)
""",
)

CLASSIC_PROMPT_DEDENT_SINGLE_LINE = (
    ">>>     print(1)\n",
    "print(1)\n",
)

CLASSIC_PROMPT_DEDENT_LEADING_WS = (
    "    >>>     print(1)\n",
    "print(1)\n",
)

CLASSIC_PROMPT_MULTILINE_DOCTEST = (
    """\
>>> for i in range(2):
...     print(i)
""",
    """\
for i in range(2):
    print(i)
""",
)

CLASSIC_PROMPT_STANDALONE_CONTINUATION = (
    "...     print(1)\n",
    "...     print(1)\n",
)


def test_classic_prompt():
    for sample, expected in [
        CLASSIC_PROMPT,
        CLASSIC_PROMPT_L2,
        CLASSIC_PROMPT_L3,
        CLASSIC_PROMPT_DEDENT_SINGLE_LINE,
        CLASSIC_PROMPT_DEDENT_LEADING_WS,
        CLASSIC_PROMPT_MULTILINE_DOCTEST,
        CLASSIC_PROMPT_STANDALONE_CONTINUATION,
    ]:
        assert ipt2.classic_prompt(
            sample.splitlines(keepends=True)
        ) == expected.splitlines(keepends=True)


IPYTHON_PROMPT = (
    """\
In [1]: for a in range(5):
   ...:     print(a)
""",
    """\
for a in range(5):
    print(a)
""",
)

IPYTHON_PROMPT_L2 = (
    """\
for a in range(5):
   ...:     print(a)
   ...:     print(a ** 2)
""",
    """\
for a in range(5):
    print(a)
    print(a ** 2)
""",
)


IPYTHON_PROMPT_VI_INS = (
    """\
[ins] In [11]: def a():
          ...:     123
          ...:
          ...: 123
""",
    """\
def a():
    123

123
""",
)

IPYTHON_PROMPT_VI_NAV = (
    """\
[nav] In [11]: def a():
          ...:     123
          ...:
          ...: 123
""",
    """\
def a():
    123

123
""",
)


def test_ipython_prompt():
    for sample, expected in [
        IPYTHON_PROMPT,
        IPYTHON_PROMPT_L2,
        IPYTHON_PROMPT_VI_INS,
        IPYTHON_PROMPT_VI_NAV,
    ]:
        assert ipt2.ipython_prompt(
            sample.splitlines(keepends=True)
        ) == expected.splitlines(keepends=True)


INDENT_SPACES = (
    """\
     if True:
        a = 3
""",
    """\
if True:
   a = 3
""",
)

INDENT_TABS = (
    """\
\tif True:
\t\tb = 4
""",
    """\
if True:
\tb = 4
""",
)


def test_leading_indent():
    for sample, expected in [INDENT_SPACES, INDENT_TABS]:
        assert ipt2.leading_indent(
            sample.splitlines(keepends=True)
        ) == expected.splitlines(keepends=True)


INDENT_SPACES_COMMENT = (
    """\
    # comment
if True:
    a = 3
""",
    """\
    # comment
if True:
    a = 3
""",
)

INDENT_TABS_COMMENT = (
    """\
\t# comment
if True:
\tb = 4
""",
    """\
\t# comment
if True:
\tb = 4
""",
)


INDENTED_CODE_WITH_ALIGNED_COMMENT = (
    """\
    # comment
    x = 1
    print(x)
""",
    """\
# comment
x = 1
print(x)
""",
)


@pytest.mark.parametrize(
    "sample, expected",
    [INDENT_SPACES_COMMENT, INDENT_TABS_COMMENT, INDENTED_CODE_WITH_ALIGNED_COMMENT],
)
def test_leading_indent_comment(sample, expected):
    assert ipt2.leading_indent(sample.splitlines(keepends=True)) == expected.splitlines(
        keepends=True
    )


LEADING_EMPTY_LINES = (
    """\
    \t

if True:
    a = 3

b = 4
""",
    """\
if True:
    a = 3

b = 4
""",
)

ONLY_EMPTY_LINES = (
    """\
    \t

""",
    """\
    \t

""",
)


def test_leading_empty_lines():
    for sample, expected in [LEADING_EMPTY_LINES, ONLY_EMPTY_LINES]:
        assert ipt2.leading_empty_lines(
            sample.splitlines(keepends=True)
        ) == expected.splitlines(keepends=True)


CRLF_MAGIC = (["%%ls\r\n"], ["get_ipython().run_cell_magic('ls', '', '')\n"])


def test_crlf_magic():
    for sample, expected in [CRLF_MAGIC]:
        assert ipt2.cell_magic(sample) == expected



"""Error script. DO NOT EDIT FURTHER! It will break exception doctests!!!"""

import sys


def div0():
    "foo"
    x = 1
    y = 0
    x / y


def sysexit(stat, mode):
    raise SystemExit(stat, f"Mode = {mode}")


def bar(mode):
    "bar"
    if mode == "div":
        div0()
    elif mode == "exit":
        try:
            stat = int(sys.argv[2])
        except:
            stat = 1
        sysexit(stat, mode)
    else:
        raise ValueError("Unknown mode")


if __name__ == "__main__":
    try:
        mode = sys.argv[1]
    except IndexError:
        mode = "div"

    bar(mode)



import sys
from IPython.testing.tools import AssertPrints, AssertNotPrints
from IPython.core.displayhook import CapturingDisplayHook
from IPython.utils.capture import CapturedIO


def test_output_displayed():
    """Checking to make sure that output is displayed"""

    with AssertPrints("2"):
        ip.run_cell("1+1", store_history=True)

    with AssertPrints("2"):
        ip.run_cell("1+1 # comment with a semicolon;", store_history=True)

    with AssertPrints("2"):
        ip.run_cell("1+1\n#commented_out_function();", store_history=True)


def test_output_quiet():
    """Checking to make sure that output is quiet"""

    with AssertNotPrints("2"):
        ip.run_cell("1+1;", store_history=True)

    with AssertNotPrints("2"):
        ip.run_cell("1+1; # comment with a semicolon", store_history=True)

    with AssertNotPrints("2"):
        ip.run_cell("1+1;\n#commented_out_function()", store_history=True)


def test_underscore_no_overwrite_user():
    ip.run_cell("_ = 42", store_history=True)
    ip.run_cell("1+1", store_history=True)

    with AssertPrints("42"):
        ip.run_cell("print(_)", store_history=True)

    ip.run_cell("del _", store_history=True)
    ip.run_cell("6+6", store_history=True)
    with AssertPrints("12"):
        ip.run_cell("_", store_history=True)


def test_underscore_no_overwrite_builtins():
    ip.run_cell("import gettext ; gettext.install('foo')", store_history=True)
    ip.run_cell("3+3", store_history=True)

    with AssertPrints("gettext"):
        ip.run_cell("print(_)", store_history=True)

    ip.run_cell('_ = "userset"', store_history=True)

    with AssertPrints("userset"):
        ip.run_cell("print(_)", store_history=True)
    ip.run_cell("import builtins; del builtins._")


def test_interactivehooks_ast_modes():
    """
    Test that ast nodes can be triggered with different modes
    """
    saved_mode = ip.ast_node_interactivity
    ip.ast_node_interactivity = "last_expr_or_assign"

    try:
        with AssertPrints("2"):
            ip.run_cell("a = 1+1", store_history=True)

        with AssertPrints("9"):
            ip.run_cell("b = 1+8 # comment with a semicolon;", store_history=False)

        with AssertPrints("7"):
            ip.run_cell("c = 1+6\n#commented_out_function();", store_history=True)

        ip.run_cell("d = 11", store_history=True)
        with AssertPrints("12"):
            ip.run_cell("d += 1", store_history=True)

        with AssertNotPrints("42"):
            ip.run_cell("(u,v) = (41+1, 43-1)")

    finally:
        ip.ast_node_interactivity = saved_mode


def test_interactivehooks_ast_modes_semi_suppress():
    """
    Test that ast nodes can be triggered with different modes and suppressed
    by semicolon
    """
    saved_mode = ip.ast_node_interactivity
    ip.ast_node_interactivity = "last_expr_or_assign"

    try:
        with AssertNotPrints("2"):
            ip.run_cell("x = 1+1;", store_history=True)

        with AssertNotPrints("7"):
            ip.run_cell("y = 1+6; # comment with a semicolon", store_history=True)

        with AssertNotPrints("9"):
            ip.run_cell("z = 1+8;\n#commented_out_function()", store_history=True)

    finally:
        ip.ast_node_interactivity = saved_mode


def test_capture_display_hook_format():
    """Tests that the capture display hook conforms to the CapturedIO output format"""
    hook = CapturingDisplayHook(ip)
    hook({"foo": "bar"})
    captured = CapturedIO(sys.stdout, sys.stderr, hook.outputs)
    # Should not raise with RichOutput transformation error
    captured.outputs



"""Directives and roles for documenting traitlets config options.

::

    .. configtrait:: Application.log_datefmt

        Description goes here.

    Cross reference like this: :configtrait:`Application.log_datefmt`.
"""


def setup(app):
    app.add_object_type("configtrait", "configtrait", objname="Config option")
    metadata = {"parallel_read_safe": True, "parallel_write_safe": True}
    return metadata



# encoding: utf-8
"""Tests for IPython.utils.capture"""

# -----------------------------------------------------------------------------
#  Copyright (C) 2013 The IPython Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file COPYING, distributed as part of this software.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------


import sys

import pytest

from IPython.utils import capture

# -----------------------------------------------------------------------------
# Globals
# -----------------------------------------------------------------------------

_mime_map = dict(
    _repr_png_="image/png",
    _repr_jpeg_="image/jpeg",
    _repr_svg_="image/svg+xml",
    _repr_html_="text/html",
    _repr_json_="application/json",
    _repr_javascript_="application/javascript",
)

basic_data = {
    "image/png": b"binarydata",
    "text/html": "<b>bold</b>",
}
basic_metadata = {
    "image/png": {
        "width": 10,
        "height": 20,
    },
}

full_data = {
    "image/png": b"binarydata",
    "image/jpeg": b"binarydata",
    "image/svg+xml": "<svg>",
    "text/html": "<b>bold</b>",
    "application/javascript": "alert();",
    "application/json": "{}",
}
full_metadata = {
    "image/png": {"png": "exists"},
    "image/jpeg": {"jpeg": "exists"},
    "image/svg+xml": {"svg": "exists"},
    "text/html": {"html": "exists"},
    "application/javascript": {"js": "exists"},
    "application/json": {"json": "exists"},
}

hello_stdout = "hello, stdout"
hello_stderr = "hello, stderr"


# -----------------------------------------------------------------------------
# Test Functions
# -----------------------------------------------------------------------------
@pytest.mark.parametrize("method_mime", _mime_map.items())
def test_rich_output_empty(method_mime):
    """RichOutput with no args"""
    rich = capture.RichOutput()
    method, mime = method_mime
    assert getattr(rich, method)() is None


def test_rich_output():
    """test RichOutput basics"""
    data = basic_data
    metadata = basic_metadata
    rich = capture.RichOutput(data=data, metadata=metadata)
    assert rich._repr_html_() == data["text/html"]
    assert rich._repr_png_() == (data["image/png"], metadata["image/png"])
    assert rich._repr_latex_() is None
    assert rich._repr_javascript_() is None
    assert rich._repr_svg_() is None


@pytest.mark.parametrize("method_mime", _mime_map.items())
def test_rich_output_no_metadata(method_mime):
    """test RichOutput with no metadata"""
    data = full_data
    rich = capture.RichOutput(data=data)
    method, mime = method_mime
    assert getattr(rich, method)() == data[mime]


@pytest.mark.parametrize("method_mime", _mime_map.items())
def test_rich_output_metadata(method_mime):
    """test RichOutput with metadata"""
    data = full_data
    metadata = full_metadata
    rich = capture.RichOutput(data=data, metadata=metadata)
    method, mime = method_mime
    assert getattr(rich, method)() == (data[mime], metadata[mime])


def test_rich_output_display():
    """test RichOutput.display

    This is a bit circular, because we are actually using the capture code we are testing
    to test itself.
    """
    data = full_data
    rich = capture.RichOutput(data=data)
    with capture.capture_output() as cap:
        rich.display()
    assert len(cap.outputs) == 1
    rich2 = cap.outputs[0]
    assert rich2.data == rich.data
    assert rich2.metadata == rich.metadata


def test_capture_output():
    """capture_output works"""
    rich = capture.RichOutput(data=full_data)
    with capture.capture_output() as cap:
        print(hello_stdout, end="")
        print(hello_stderr, end="", file=sys.stderr)
        rich.display()
    assert hello_stdout == cap.stdout
    assert hello_stderr == cap.stderr


def test_capture_output_no_stdout():
    """test capture_output(stdout=False)"""
    rich = capture.RichOutput(data=full_data)
    with capture.capture_output(stdout=False) as cap:
        print(hello_stdout, end="")
        print(hello_stderr, end="", file=sys.stderr)
        rich.display()
    assert "" == cap.stdout
    assert hello_stderr == cap.stderr
    assert len(cap.outputs) == 1


def test_capture_output_no_stderr():
    """test capture_output(stderr=False)"""
    rich = capture.RichOutput(data=full_data)
    # add nested capture_output so stderr doesn't make it to nose output
    with capture.capture_output(), capture.capture_output(stderr=False) as cap:
        print(hello_stdout, end="")
        print(hello_stderr, end="", file=sys.stderr)
        rich.display()
    assert hello_stdout == cap.stdout
    assert "" == cap.stderr
    assert len(cap.outputs) == 1


def test_capture_output_no_display():
    """test capture_output(display=False)"""
    rich = capture.RichOutput(data=full_data)
    with capture.capture_output(display=False) as cap:
        print(hello_stdout, end="")
        print(hello_stderr, end="", file=sys.stderr)
        rich.display()
    assert hello_stdout == cap.stdout
    assert hello_stderr == cap.stderr
    assert cap.outputs == []



# -*- coding: utf-8 -*-
"""Setup script for IPython.

Under Posix environments it works like a typical setup.py script.
Under Windows, the command sdist is not supported, since IPython
requires utilities which are not available under Windows."""

#-----------------------------------------------------------------------------
#  Copyright (c) 2008-2011, IPython Development Team.
#  Copyright (c) 2001-2007, Fernando Perez <fernando.perez@colorado.edu>
#  Copyright (c) 2001, Janko Hauser <jhauser@zscout.de>
#  Copyright (c) 2001, Nathaniel Gray <n8gray@caltech.edu>
#
#  Distributed under the terms of the Modified BSD License.
#
#  The full license is in the file COPYING.rst, distributed with this software.
#-----------------------------------------------------------------------------

import os
import sys

# At least we're on the python version we need, move on.

from setuptools import setup

# Our own imports

from setupbase import target_update

from setupbase import (
    setup_args,
    check_package_data_first,
    find_data_files,
    git_prebuild,
)

#-------------------------------------------------------------------------------
# Things related to the IPython documentation
#-------------------------------------------------------------------------------

# update the manuals when building a source dist
if len(sys.argv) >= 2 and sys.argv[1] in ('sdist','bdist_rpm'):

    # List of things to be updated. Each entry is a triplet of args for
    # target_update()
    to_update = [
        (
            "docs/man/ipython.1.gz",
            ["docs/man/ipython.1"],
            "cd docs/man && python -m gzip --best ipython.1",
        ),
    ]


    [ target_update(*t) for t in to_update ]

#---------------------------------------------------------------------------
# Find all the packages, package data, and data_files
#---------------------------------------------------------------------------

data_files = find_data_files()

setup_args['data_files'] = data_files

#---------------------------------------------------------------------------
# custom distutils commands
#---------------------------------------------------------------------------
# imports here, so they are after setuptools import if there was one
from setuptools.command.sdist import sdist

setup_args['cmdclass'] = {
    'build_py': \
            check_package_data_first(git_prebuild('IPython')),
    'sdist' : git_prebuild('IPython', sdist),
}

#---------------------------------------------------------------------------
# Do the actual setup now
#---------------------------------------------------------------------------

if __name__ == "__main__":
    setup(**setup_args)



# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import json
import os
import warnings

from unittest import mock

import pytest

from IPython import display
from IPython.core.getipython import get_ipython
from IPython.utils.io import capture_output
from IPython.utils.tempdir import NamedFileInTemporaryDirectory
from IPython import paths as ipath
from IPython.testing.tools import AssertNotPrints

import IPython.testing.decorators as dec


def test_image_size():
    """Simple test for display.Image(args, width=x,height=y)"""
    thisurl = "http://www.google.fr/images/srpr/logo3w.png"
    img = display.Image(url=thisurl, width=200, height=200)
    assert '<img src="%s" width="200" height="200"/>' % (thisurl) == img._repr_html_()
    img = display.Image(url=thisurl, metadata={"width": 200, "height": 200})
    assert '<img src="%s" width="200" height="200"/>' % (thisurl) == img._repr_html_()
    img = display.Image(url=thisurl, width=200)
    assert '<img src="%s" width="200"/>' % (thisurl) == img._repr_html_()
    img = display.Image(url=thisurl)
    assert '<img src="%s"/>' % (thisurl) == img._repr_html_()
    img = display.Image(url=thisurl, unconfined=True)
    assert '<img src="%s" class="unconfined"/>' % (thisurl) == img._repr_html_()


def test_image_mimes():
    fmt = get_ipython().display_formatter.format
    for format in display.Image._ACCEPTABLE_EMBEDDINGS:
        mime = display.Image._MIMETYPES[format]
        img = display.Image(b"garbage", format=format)
        data, metadata = fmt(img)
        assert sorted(data) == sorted([mime, "text/plain"])


def test_geojson():
    gj = display.GeoJSON(
        data={
            "type": "Feature",
            "geometry": {"type": "Point", "coordinates": [-81.327, 296.038]},
            "properties": {"name": "Inca City"},
        },
        url_template="http://s3-eu-west-1.amazonaws.com/whereonmars.cartodb.net/{basemap_id}/{z}/{x}/{y}.png",
        layer_options={
            "basemap_id": "celestia_mars-shaded-16k_global",
            "attribution": "Celestia/praesepe",
            "minZoom": 0,
            "maxZoom": 18,
        },
    )
    assert "<IPython.core.display.GeoJSON object>" == str(gj)


def test_retina_png():
    here = os.path.dirname(__file__)
    img = display.Image(os.path.join(here, "2x2.png"), retina=True)
    assert img.height == 1
    assert img.width == 1
    data, md = img._repr_png_()
    assert md["width"] == 1
    assert md["height"] == 1


def test_embed_svg_url():
    import gzip
    from io import BytesIO

    svg_data = b'<svg><circle x="0" y="0" r="1"/></svg>'
    url = "http://test.com/circle.svg"

    gzip_svg = BytesIO()
    with gzip.open(gzip_svg, "wb") as fp:
        fp.write(svg_data)
    gzip_svg = gzip_svg.getvalue()

    def mocked_urlopen(*args, **kwargs):
        class MockResponse:
            def __init__(self, svg):
                self._svg_data = svg
                self.headers = {"content-type": "image/svg+xml"}

            def read(self):
                return self._svg_data

        if args[0] == url:
            return MockResponse(svg_data)
        elif args[0] == url + "z":
            ret = MockResponse(gzip_svg)
            ret.headers["content-encoding"] = "gzip"
            return ret
        return MockResponse(None)

    with mock.patch("urllib.request.urlopen", side_effect=mocked_urlopen):
        svg = display.SVG(url=url)
        assert svg._repr_svg_().startswith("<svg") is True
        svg = display.SVG(url=url + "z")
        assert svg._repr_svg_().startswith("<svg") is True


def test_retina_jpeg():
    here = os.path.dirname(__file__)
    img = display.Image(os.path.join(here, "2x2.jpg"), retina=True)
    assert img.height == 1
    assert img.width == 1
    data, md = img._repr_jpeg_()
    assert md["width"] == 1
    assert md["height"] == 1


def test_base64image():
    display.Image(
        "iVBORw0KGgoAAAANSUhEUgAAAAEAAAABAQMAAAAl21bKAAAAA1BMVEUAAACnej3aAAAAAWJLR0QAiAUdSAAAAAlwSFlzAAALEwAACxMBAJqcGAAAAAd0SU1FB94BCRQnOqNu0b4AAAAKSURBVAjXY2AAAAACAAHiIbwzAAAAAElFTkSuQmCC"
    )


def test_image_filename_defaults():
    """test format constraint, and validity of jpeg and png"""
    tpath = ipath.get_ipython_package_dir()
    pytest.raises(
        ValueError,
        display.Image,
        filename=os.path.join(tpath, "testing/tests/badformat.zip"),
        embed=True,
    )
    pytest.raises(ValueError, display.Image)
    pytest.raises(
        ValueError,
        display.Image,
        data="this is not an image",
        format="badformat",
        embed=True,
    )


def _get_inline_config():
    from matplotlib_inline.config import InlineBackend

    return InlineBackend.instance()


@dec.skip_without("matplotlib")
def test_set_matplotlib_close():
    from matplotlib_inline.backend_inline import set_matplotlib_close

    cfg = _get_inline_config()
    cfg.close_figures = False

    set_matplotlib_close()
    assert cfg.close_figures
    set_matplotlib_close(False)
    assert not cfg.close_figures


_fmt_mime_map = {
    "png": "image/png",
    "jpeg": "image/jpeg",
    "pdf": "application/pdf",
    "retina": "image/png",
    "svg": "image/svg+xml",
}


@dec.skip_without("matplotlib")
def test_set_matplotlib_formats():
    from matplotlib.figure import Figure
    from matplotlib_inline.backend_inline import set_matplotlib_formats

    formatters = get_ipython().display_formatter.formatters
    for formats in [
        ("png",),
        ("pdf", "svg"),
        ("jpeg", "retina", "png"),
        (),
    ]:
        active_mimes = {_fmt_mime_map[fmt] for fmt in formats}
        set_matplotlib_formats(*formats)
        for mime, f in formatters.items():
            if mime in active_mimes:
                assert Figure in f
            else:
                assert Figure not in f


@dec.skip_without("matplotlib")
def test_set_matplotlib_formats_kwargs():
    from matplotlib.figure import Figure
    from matplotlib_inline.backend_inline import set_matplotlib_formats

    ip = get_ipython()
    cfg = _get_inline_config()
    cfg.print_figure_kwargs.update(dict(foo="bar"))
    kwargs = dict(dpi=150)
    try:
        set_matplotlib_formats("png", **kwargs)
        formatter = ip.display_formatter.formatters["image/png"]
        f = formatter.lookup_by_type(Figure)
        formatter_kwargs = f.keywords
        expected = kwargs
        expected["base64"] = True
        expected["fmt"] = "png"
        expected.update(cfg.print_figure_kwargs)
        assert formatter_kwargs == expected
    finally:
        cfg.print_figure_kwargs.clear()


@dec.skip_without("matplotlib")
def test_matplotlib_positioning():
    _ip = get_ipython()

    prev_active_types = _ip.display_formatter.active_types.copy()
    prev_execution_count = _ip.execution_count
    prev_user_ns_underscore = _ip.user_ns.get("_", None)

    _ip.history_manager.reset()
    _ip.display_formatter.active_types = ["text/plain", "image/png"]

    _ip.run_cell("import matplotlib")
    prev_mpl_backend = _ip.run_cell("matplotlib.get_backend()").result

    try:
        _ip.run_line_magic("matplotlib", "inline")
        _ip.execution_count = 1
        _ip.run_cell("'no plot here'", store_history=True)

        # Cell 2: No manual flush
        _ip.run_cell(
            "import matplotlib.pyplot as plt;plt.plot([0, 1])", store_history=True
        )

        _ip.run_cell("'no plot here'", store_history=True)

        # Cell 4: Manual flush
        _ip.run_cell("plt.plot([1, 0])\nplt.show()", store_history=True)

        _ip.run_cell("'no plot here'", store_history=True)

        outputs = _ip.history_manager.outputs

        # Only cells 2 and 4 should have plots
        for cell_num in [1, 3, 5]:
            assert not any(
                "image/png" in out.bundle for out in outputs.get(cell_num, [])
            ), f"Cell {cell_num} should not have plot"

        cell_2_has_plot = any("image/png" in out.bundle for out in outputs.get(2, []))
        cell_4_has_plot = any("image/png" in out.bundle for out in outputs.get(4, []))

        assert cell_2_has_plot, "Cell 2 should have plot (auto-flush)"
        assert cell_4_has_plot, "Cell 4 should have plot (manual flush)"

    finally:
        _ip.run_cell("plt.close('all')")
        _ip.run_line_magic("matplotlib", prev_mpl_backend)
        _ip.history_manager.reset()
        _ip.display_formatter.active_types = prev_active_types
        _ip.displayhook.flush()

        _ip.execution_count = prev_execution_count
        if prev_user_ns_underscore is not None:
            _ip.user_ns["_"] = prev_user_ns_underscore
        elif "_" in _ip.user_ns:
            del _ip.user_ns["_"]


def test_display_available():
    """
    Test that display is available without import

    We don't really care if it's in builtin or anything else, but it should
    always be available.
    """
    ip = get_ipython()
    with AssertNotPrints("NameError"):
        ip.run_cell("display")
    try:
        ip.run_cell("del display")
    except NameError:
        pass  # it's ok, it might be in builtins
    # even if deleted it should be back
    with AssertNotPrints("NameError"):
        ip.run_cell("display")


def test_textdisplayobj_pretty_repr():
    p = display.Pretty("This is a simple test")
    assert repr(p) == "<IPython.core.display.Pretty object>"
    assert p.data == "This is a simple test"

    p._show_mem_addr = True
    assert repr(p) == object.__repr__(p)


def test_displayobject_repr():
    h = display.HTML("<br />")
    assert repr(h) == "<IPython.core.display.HTML object>"
    h._show_mem_addr = True
    assert repr(h) == object.__repr__(h)
    h._show_mem_addr = False
    assert repr(h) == "<IPython.core.display.HTML object>"

    j = display.Javascript("")
    assert repr(j) == "<IPython.core.display.Javascript object>"
    j._show_mem_addr = True
    assert repr(j) == object.__repr__(j)
    j._show_mem_addr = False
    assert repr(j) == "<IPython.core.display.Javascript object>"


@mock.patch("warnings.warn")
def test_encourage_iframe_over_html(m_warn):
    display.HTML()
    m_warn.assert_not_called()

    display.HTML("<br />")
    m_warn.assert_not_called()

    display.HTML(
        '<html><p>Lots of content here</p><iframe src="http://a.com"></iframe>'
    )
    m_warn.assert_not_called()

    display.HTML('<iframe src="http://a.com"></iframe>')
    m_warn.assert_called_with("Consider using IPython.display.IFrame instead")

    m_warn.reset_mock()
    display.HTML('<IFRAME SRC="http://a.com"></IFRAME>')
    m_warn.assert_called_with("Consider using IPython.display.IFrame instead")


def test_progress():
    p = display.ProgressBar(10)
    assert "0/10" in repr(p)
    p.html_width = "100%"
    p.progress = 5
    assert (
        p._repr_html_() == "<progress style='width:100%' max='10' value='5'></progress>"
    )


def test_progress_iter():
    with capture_output(display=False) as captured:
        for i in display.ProgressBar(5):
            out = captured.stdout
            assert "{0}/5".format(i) in out
    out = captured.stdout
    assert "5/5" in out


def test_json():
    d = {"a": 5}
    lis = [d]
    metadata = [
        {"expanded": False, "root": "root"},
        {"expanded": True, "root": "root"},
        {"expanded": False, "root": "custom"},
        {"expanded": True, "root": "custom"},
    ]
    json_objs = [
        display.JSON(d),
        display.JSON(d, expanded=True),
        display.JSON(d, root="custom"),
        display.JSON(d, expanded=True, root="custom"),
    ]
    for j, md in zip(json_objs, metadata):
        assert j._repr_json_() == (d, md)

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        j = display.JSON(json.dumps(d))
        assert len(w) == 1
        assert j._repr_json_() == (d, metadata[0])

    json_objs = [
        display.JSON(lis),
        display.JSON(lis, expanded=True),
        display.JSON(lis, root="custom"),
        display.JSON(lis, expanded=True, root="custom"),
    ]
    for j, md in zip(json_objs, metadata):
        assert j._repr_json_() == (lis, md)

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        j = display.JSON(json.dumps(lis))
        assert len(w) == 1
        assert j._repr_json_() == (lis, metadata[0])


def test_video_embedding():
    """use a tempfile, with dummy-data, to ensure that video embedding doesn't crash"""
    v = display.Video("http://ignored")
    assert not v.embed
    html = v._repr_html_()
    assert 'src="data:' not in html
    assert 'src="http://ignored"' in html

    with pytest.raises(ValueError):
        v = display.Video(b"abc")

    with NamedFileInTemporaryDirectory("test.mp4", "wb") as f:
        f.write(b"abc")
        f.close()

        v = display.Video(f.name)
        assert not v.embed
        html = v._repr_html_()
        assert 'src="data:' not in html

        v = display.Video(f.name, embed=True)
        html = v._repr_html_()
        assert 'src="data:video/mp4;base64,YWJj"' in html

        v = display.Video(f.name, embed=True, mimetype="video/other")
        html = v._repr_html_()
        assert 'src="data:video/other;base64,YWJj"' in html

        v = display.Video(b"abc", embed=True, mimetype="video/mp4")
        html = v._repr_html_()
        assert 'src="data:video/mp4;base64,YWJj"' in html

        v = display.Video("YWJj", embed=True, mimetype="video/xyz")
        html = v._repr_html_()
        assert 'src="data:video/xyz;base64,YWJj"' in html


def test_html_metadata():
    s = "<h1>Test</h1>"
    h = display.HTML(s, metadata={"isolated": True})
    assert h._repr_html_() == (s, {"isolated": True})


def test_display_id():
    ip = get_ipython()
    with mock.patch.object(ip.display_pub, "publish") as pub:
        handle = display.display("x")
        assert handle is None
        handle = display.display("y", display_id="secret")
        assert isinstance(handle, display.DisplayHandle)
        handle2 = display.display("z", display_id=True)
        assert isinstance(handle2, display.DisplayHandle)
    assert handle.display_id != handle2.display_id

    assert pub.call_count == 3
    args, kwargs = pub.call_args_list[0]
    assert args == ()
    assert kwargs == {
        "data": {"text/plain": repr("x")},
        "metadata": {},
    }
    args, kwargs = pub.call_args_list[1]
    assert args == ()
    assert kwargs == {
        "data": {"text/plain": repr("y")},
        "metadata": {},
        "transient": {
            "display_id": handle.display_id,
        },
    }
    args, kwargs = pub.call_args_list[2]
    assert args == ()
    assert kwargs == {
        "data": {"text/plain": repr("z")},
        "metadata": {},
        "transient": {
            "display_id": handle2.display_id,
        },
    }


def test_update_display():
    ip = get_ipython()
    with mock.patch.object(ip.display_pub, "publish") as pub:
        with pytest.raises(TypeError):
            display.update_display("x")
        display.update_display("x", display_id="1")
        display.update_display("y", display_id="2")
    args, kwargs = pub.call_args_list[0]
    assert args == ()
    assert kwargs == {
        "data": {"text/plain": repr("x")},
        "metadata": {},
        "transient": {
            "display_id": "1",
        },
        "update": True,
    }
    args, kwargs = pub.call_args_list[1]
    assert args == ()
    assert kwargs == {
        "data": {"text/plain": repr("y")},
        "metadata": {},
        "transient": {
            "display_id": "2",
        },
        "update": True,
    }


def test_display_handle():
    ip = get_ipython()
    handle = display.DisplayHandle()
    assert isinstance(handle.display_id, str)
    handle = display.DisplayHandle("my-id")
    assert handle.display_id == "my-id"
    with mock.patch.object(ip.display_pub, "publish") as pub:
        handle.display("x")
        handle.update("y")

    args, kwargs = pub.call_args_list[0]
    assert args == ()
    assert kwargs == {
        "data": {"text/plain": repr("x")},
        "metadata": {},
        "transient": {
            "display_id": handle.display_id,
        },
    }
    args, kwargs = pub.call_args_list[1]
    assert args == ()
    assert kwargs == {
        "data": {"text/plain": repr("y")},
        "metadata": {},
        "transient": {
            "display_id": handle.display_id,
        },
        "update": True,
    }


def test_image_alt_tag():
    """Simple test for display.Image(args, alt=x,)"""
    thisurl = "http://example.com/image.png"
    img = display.Image(url=thisurl, alt="an image")
    assert '<img src="%s" alt="an image"/>' % (thisurl) == img._repr_html_()
    img = display.Image(url=thisurl, unconfined=True, alt="an image")
    assert (
        '<img src="%s" class="unconfined" alt="an image"/>' % (thisurl)
        == img._repr_html_()
    )
    img = display.Image(url=thisurl, alt='>"& <')
    assert '<img src="%s" alt="&gt;&quot;&amp; &lt;"/>' % (thisurl) == img._repr_html_()

    img = display.Image(url=thisurl, metadata={"alt": "an image"})
    assert img.alt == "an image"
    here = os.path.dirname(__file__)
    img = display.Image(os.path.join(here, "2x2.png"), alt="an image")
    assert img.alt == "an image"
    _, md = img._repr_png_()
    assert md["alt"] == "an image"


def test_image_bad_filename_raises_proper_exception():
    with pytest.raises(FileNotFoundError):
        display.Image("/this/file/does/not/exist/")._repr_png_()



#!/usr/bin/env python
from __future__ import annotations
import ast
import os
import platform
import pytest
import random
import shutil
import sys
import tempfile
import textwrap
import time
import unittest
from types import ModuleType

from IPython.extensions.autoreload import AutoreloadMagics

from IPython.extensions.deduperreload.deduperreload import compare_ast, DeduperReloader

if platform.python_implementation() != "CPython":
    pytest.skip(
        "We do not support non-CPython versions.",
        allow_module_level=True,
    )


class DeduperTestReloader(DeduperReloader):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.exceptions_raised: list[Exception] = []

    def _patch_namespace(
        self, module: ModuleType | type, prefixes: list[str] | None = None
    ) -> bool:
        try:
            assert super()._patch_namespace(module, prefixes)
            return True
        except Exception as e:
            self.exceptions_raised.append(e)
            return False


def is_nonempty_file(fname: str) -> bool:
    return len(fname) > 0


def squish_text(text: str) -> str:
    """
    Turns text like this:

    '''        def foo():
    return "bar"
            def baz():
                return "bat"
    def bam():
                return "bat"
    '''

    into this:

    '''def foo():
        return "bar"
    def baz():
        return "bat"
    def bam():
        return "bat"
    '''

    The former is common when we are trying to use string templates
    whose parameters are multiline and unaware of the existing indentation.

    :param text: a string with messed up indentation
    :return: `text` but with indentation fixed
    """
    prev_indentation = 0
    transformed_text_lines = []
    for line in text.strip("\n").splitlines():
        line_without_indentation = line.lstrip()
        indentation = len(line) - len(line_without_indentation)
        if indentation == 0:
            indentation = prev_indentation
        else:
            prev_indentation = indentation
        transformed_text_lines.append(
            textwrap.indent(line_without_indentation, " " * indentation)
        )
    return textwrap.dedent("\n".join(transformed_text_lines))


class AutoreloadDetectionSuite(unittest.TestCase):
    """
    Unit tests for autoreload testing logic
    """

    def test_compare_ast(self):
        code1 = squish_text(
            """
            def factorial(n):
                def fn(sdfsdf):
                    print(sdfsdf)
                if n < 0:
                    return "Factorial is not defined for negative numbers."
                elif n == 0 or n == 1:
                    return 1
                else:
                    result = 1
                    for i in range(2, n + 1):
                        result *= i
                    return result
                y = 12
                print(y)
            print(1)
            x = 1212121
        """
        )
        code2 = squish_text(
            """
            def factorial(n):
                def fn(sdfsdf):
                    print(sdfsdf+"!!!")
                if n < 0:
                    return "Factorial is not defined for negative numbers."
                elif n == 0 or n == 1:
                    return 1
                else:
                    result = 1
                    for i in range(2, n + 1):
                        result *= i
                    return result
                y = 12
                print(y)
            print(1)
            x = 1212121
        """
        )
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert not compare_ast(ast_1, ast_2)

    def test_compare_ast2(self):
        code1 = squish_text(
            """
            def factorial(n):
                def fn(sdfsdf):
                    print(sdfsdf)
                if n < 0:
                    return "Factorial is not defined for negative numbers."
                elif n == 0 or n == 1:
                    return 1
                else:
                    result = 1
                    for i in range(2, n + 1):
                        result *= i
                    return result
                y = 12
                print(y)
            print(1)
            x = 1212121
        """
        )
        code2 = squish_text(
            """
            def factorial(n):
                def fn(sdfsdf):
                    print(sdfsdf)
                if n < 0:
                    return "Factorial is not defined for negative numbers."
                elif n == 0 or n == 1:
                    return 1
                else:
                    result = 1
                    for i in range(2, n + 1):
                        result *= i
                    return result
                y = 12
                print(y)
            print(1)
            x = 1212121
        """
        )
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert compare_ast(ast_1, ast_2)

    def test_autoreload_no_changes(self):
        code1 = squish_text(
            """
            def factorial(n):
                print(n)
            print(1)
            x = 1212121
        """
        )
        code2 = squish_text(
            """
            def factorial(n):
                print(n)
            print(1)
            x = 1212121
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)
        assert deduperreloader._to_autoreload.defs_to_reload == []

    def test_autoreload_static_assign_change_outside_function(self):
        code1 = squish_text(
            """
            def factorial(n):
                print(n)
            print(1)
        """
        )
        code2 = squish_text(
            """
            def factorial(n):
                print(n)
            print(1)
            x = 1212121
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)

    def test_autoreload_changes_inside_function(self):
        code1 = squish_text(
            """
            def factorial(n):
                print(n)
            print(1)
        """
        )
        code2 = squish_text(
            """
            def factorial(n):
                print(n + "edit!")
            print(1)
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)
        assert len(deduperreloader._to_autoreload.defs_to_reload) == 1

    def test_autoreload_changes_inside_and_outside_function(self):
        code1 = squish_text(
            """
            def factorial(n):
                print(n)
            print(1)
        """
        )
        code2 = squish_text(
            """
            def factorial(n):
                print(n + "edit!")
            print(2)
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert not deduperreloader.detect_autoreload(ast_1, ast_2)

    def test_autoreload_changes_inner_function(self):
        code1 = squish_text(
            """
            def factorial(n):
                def foo():
                    x = 3   
                    return x
                return foo
        """
        )
        code2 = squish_text(
            """
            def factorial(n):
                def foo():
                    x = 4  
                    return x
                return foo
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)
        assert len(deduperreloader._to_autoreload.defs_to_reload) == 1

    def test_autoreload_changes_multiple_function(self):
        code1 = squish_text(
            """
            def factorial(n):
                def foo():
                    x = 3   
                    return x
                return foo
            def bar():
                return 1
        """
        )
        code2 = squish_text(
            """
            def factorial(n):
                def foo():
                    x = 4  
                    return x
                return foo
            def bar():
                return 2
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)
        assert len(deduperreloader._to_autoreload.defs_to_reload) == 2

    def test_autoreload_change_one_function_of_multiple(self):
        code1 = squish_text(
            """
            def factorial(n):
                def foo():
                    x = 3   
                    return x
                return foo
            def bar():
                return 1
        """
        )
        code2 = squish_text(
            """
            def factorial(n):
                def foo():
                    x = 4  
                    return x
                return foo
            def bar():
                return 1
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)
        assert len(deduperreloader._to_autoreload.defs_to_reload) == 1

    def test_autoreload_handling_moves(self):
        code1 = squish_text(
            """
            def factorial(n):
                return 1
            x = 1
        """
        )
        code2 = squish_text(
            """
            x = 1
            def factorial(n):
                return 1
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)
        assert len(deduperreloader._to_autoreload.defs_to_reload) == 0

    def test_autoreload_handling_function_moves_success(self):
        code1 = squish_text(
            """
            def factorial(n):
                return 1
            x = 1
            def foo():
                return 23
            x = 2
        """
        )
        code2 = squish_text(
            """
        x = 1
        x = 2
        def foo():
            return 23
        def factorial(n):
            return 1
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)
        assert len(deduperreloader._to_autoreload.defs_to_reload) == 0

    def test_autoreload_handling_function_moves_only(self):
        code1 = squish_text(
            """
        def factorial(n):
            return 1
        x = 1
        x = 2
        """
        )
        code2 = squish_text(
            """
        x = 1
        x = 2
        def factorial(n):
            return 1
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)

    def test_autoreload_handles_new_imports(self):
        code1 = squish_text(
            """
        def factorial(n):
            return 1
        x = 1
        """
        )
        code2 = squish_text(
            """
        import ast
        def factorial(n):
            return 1
        x = 1
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)

    def test_autoreload_async_function(self):
        code1 = squish_text(
            """
        async def sleep():
            print(f'Time: {time.time() - start:.2f}')
            await asyncio.sleep(1)
        """
        )
        code2 = squish_text(
            """
        async def sleep():
            print(f'Time: {time.time() - start:.2f}')
            await asyncio.sleep(10)
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)
        assert len(deduperreloader._to_autoreload.defs_to_reload) == 1

    def test_autoreload_add_function(self):
        code1 = squish_text(
            """
        async def sleep():
            print(f'Time: {time.time() - start:.2f}')
            await asyncio.sleep(1)
        """
        )
        code2 = squish_text(
            """
        async def sleep():
            print(f'Time: {time.time() - start:.2f}')
            await asyncio.sleep(1)
        def add(x,y):
            pass
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)
        assert list(d[0][0] for d in deduperreloader._to_autoreload.defs_to_reload) == [
            "add"
        ]

    def test_autoreload_add_function_ellipsis(self):
        code1 = squish_text(
            """
        async def sleep():
            print(f'Time: {time.time() - start:.2f}')
            await asyncio.sleep(1)
        """
        )
        code2 = squish_text(
            """
        async def sleep():
            print(f'Time: {time.time() - start:.2f}')
            await asyncio.sleep(1)
        def add(x,y):
            ...
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)
        assert list(d[0][0] for d in deduperreloader._to_autoreload.defs_to_reload) == [
            "add"
        ]


class AutoreloadPatchingSuite(unittest.TestCase):
    """
    Unit tests for autoreload patching logic
    """

    def setUp(self) -> None:
        self.deduperreloader = DeduperTestReloader()

    def test_patching(self):
        code1 = squish_text(
            """
            def foo():
                return 1
        """
        )
        code2 = squish_text(
            """
            def foo():
                return 2
        """
        )
        self.deduperreloader._to_autoreload.defs_to_reload = [
            (("foo",), ast.parse(code2))
        ]
        mod = ModuleType("mod")
        exec(code1, mod.__dict__)
        self.deduperreloader._patch_namespace(mod)
        assert mod.foo() == 2

    def test_patching_parameters(self):
        code1 = squish_text(
            """
            def foo(n,s):
                return n+s
        """
        )
        code2 = squish_text(
            """
            def foo(n):
                return n
        """
        )
        self.deduperreloader._to_autoreload.defs_to_reload = [
            (("foo",), ast.parse(code2))
        ]
        mod = ModuleType("mod")
        exec(code1, mod.__dict__)
        self.deduperreloader._patch_namespace(mod)
        assert mod.foo(2) == 2

    def test_add_function(self):
        code1 = squish_text(
            """
            def foo2(n):
                return n
        """
        )
        code2 = squish_text(
            """
            def foo(n):
                return 55
        """
        )
        self.deduperreloader._to_autoreload.defs_to_reload = [
            (("foo",), ast.parse(code2))
        ]
        mod = ModuleType("mod")
        exec(code1, mod.__dict__)
        self.deduperreloader._patch_namespace(mod)
        assert mod.foo(2) == 55
        assert mod.foo2(2) == 2

    def test_two_operations(self):
        code1 = squish_text(
            """
            def foo(n):
                return 1
        """
        )
        code2 = squish_text(
            """
            def foo(n):
                return 55
        """
        )
        code3 = squish_text(
            """
            def goo():
                return -1
            def foo(n):
                x = 2
                return x+n
            def bar():
                return 200
        """
        )
        self.deduperreloader._to_autoreload.defs_to_reload = [
            (("foo",), ast.parse(code2))
        ]
        mod = ModuleType("mod")
        exec(code1, mod.__dict__)
        assert mod.foo(2) == 1
        self.deduperreloader._patch_namespace(mod)
        assert mod.foo(2) == 55
        self.deduperreloader._to_autoreload.defs_to_reload = [
            (("foo",), ast.parse(code3))
        ]
        self.deduperreloader._patch_namespace(mod)
        assert mod.foo(2) == 4

    def test_using_outside_param(self):
        code1 = squish_text(
            """
            x=1
            def foo(n):
                return 1
        """
        )
        code2 = squish_text(
            """
            x=1
            def foo(n):
                return 55+x
        """
        )
        self.deduperreloader._to_autoreload.defs_to_reload = [
            (("foo",), ast.parse(code2))
        ]
        mod = ModuleType("mod")
        exec(code1, mod.__dict__)
        assert mod.foo(2) == 1
        self.deduperreloader._patch_namespace(mod)
        assert mod.foo(2) == 56

    def test_importing_func(self):
        code1 = squish_text(
            """
            from os import environ
            def foo(n):
                pass
        """
        )
        code2 = squish_text(
            """
            from os import environ
            def foo():
                environ._data
                return 1
        """
        )
        self.deduperreloader._to_autoreload.defs_to_reload = [
            (("foo",), ast.parse(code2))
        ]
        mod = ModuleType("mod")
        exec(code1, mod.__dict__)
        self.deduperreloader._patch_namespace(mod)
        assert mod.foo() == 1


class FakeShell:
    def __init__(self):
        self.ns = {}
        self.user_ns = self.ns
        self.user_ns_hidden = {}
        self.auto_magics = AutoreloadMagics(shell=self)

    @staticmethod
    def pre_run_cell(obj):
        try_with_arg = False
        try:
            obj.pre_run_cell()
        except TypeError:
            try_with_arg = True
        if try_with_arg:
            obj.pre_run_cell(None)

    @staticmethod
    def post_run_cell(obj):
        try_with_arg = False
        try:
            obj.post_run_cell()
        except TypeError:
            try_with_arg = True
        if try_with_arg:
            obj.post_run_cell(None)

    def run_code(self, code):
        self.pre_run_cell(self.auto_magics)
        exec(code, self.user_ns)
        self.auto_magics.post_execute_hook()

    def push(self, items):
        self.ns.update(items)

    def magic_autoreload(self, parameter):
        self.auto_magics.autoreload(parameter)


class ShellFixture(unittest.TestCase):
    """Fixture for creating test module files"""

    test_dir = None
    old_sys_path = None
    filename_chars = "abcdefghijklmopqrstuvwxyz0123456789"

    def setUp(self):
        self.created_temp_modules = set()
        self.test_dir = tempfile.mkdtemp()
        self.old_sys_path = list(sys.path)
        sys.path.insert(0, self.test_dir)
        self.shell = FakeShell()

    def tearDown(self):
        for mod_name in self.created_temp_modules:
            sys.modules.pop(mod_name, None)
        shutil.rmtree(self.test_dir)
        sys.path = self.old_sys_path

        self.test_dir = None
        self.old_sys_path = None
        self.shell = None

    def get_module(self):
        module_name = "tmpmod_" + "".join(random.sample(self.filename_chars, 20))
        if module_name in sys.modules:
            del sys.modules[module_name]
        file_name = os.path.join(self.test_dir, module_name + ".py")
        return module_name, file_name

    def write_file(self, filename, content):
        """
        Write a file, and force a timestamp difference of at least one second

        Notes
        -----
        Python's .pyc files record the timestamp of their compilation
        with a time resolution of one second.

        Therefore, we need to force a timestamp difference between .py
        and .pyc, without having the .py file be timestamped in the
        future, and without changing the timestamp of the .pyc file
        (because that is stored in the file). The only reliable way
        to achieve this seems to be to sleep.

        Doesn't seem necessary on Darwin so we make this the exception.
        """
        if platform.system().lower() != "darwin":
            time.sleep(1.05)
        with open(filename, "w", encoding="utf-8") as f:
            f.write(squish_text(content))

    def new_module(self, code):
        mod_name, mod_fn = self.get_module()
        with open(mod_fn, "w", encoding="utf-8") as f:
            f.write(squish_text(code))
        self.created_temp_modules.add(mod_name)
        return mod_name, mod_fn


class AutoreloadHookSuite(ShellFixture):
    def test_deduperreloader_basic(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            x = 9
            def foo(y):
                return y + 3
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            x = 9
            def foo(y):
                return y + 5
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.foo(0) == 5

    def test_deduperreloader_basic2(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            x = 9
            def foo(y):
                return y + 3
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            x = 9
            def foo(y):
                return y + 5
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.foo(0) == 5

    def test_deduperreloader_basic3(self):
        mod_name, mod_fn = self.new_module(
            """
            x = 9
            def foo(y):
                return y + 3
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.magic_autoreload("2")
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            x = 9
            def foo(y):
                return y + 5
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.foo(0) == 5

    def test_deduperreloader_basic4(self):
        mod_name, mod_fn = self.new_module(
            """
            x = 9
            def foo(y):
                return y + 3
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.magic_autoreload("2")
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            x = 9
            def foo(y):
                return y + 5
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.foo(0) == 5

    def test_deduperreloader_basic5(self):
        mod_name, mod_fn = self.new_module(
            """
            x = 9
            def foo(y):
                return y + 3
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        self.shell.magic_autoreload("2")
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            x = 9
            def foo(y):
                return y + 5
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.foo(0) == 5

    def test_deduperreloader_basic6(self):
        mod_name, mod_fn = self.new_module(
            """
            x = 9
            def foo(y):
                return y + 3
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        self.shell.magic_autoreload("2")
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            x = 9
            def foo(y):
                return y + 5
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.foo(0) == 5

    def test_super(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            class Foo:
                def foo(self):
                    return 1
            class Bar(Foo):
                def bar(self):
                    return super().foo() + 1
        """
        )
        self.shell.run_code(f"from {mod_name} import Bar; bar = Bar()")
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            class Foo:
                def foo(self):
                    return 1
            class Bar(Foo):
                def bar(self):
                    return super().foo() + 2
            """,
        )
        self.shell.run_code("result = bar.bar()")
        # NOTE : only works on CPython, requires read-only patching.
        if platform.python_implementation() == "CPython":
            assert self.shell.user_ns["result"] == 3

    def test_deduperreloader_need_to_default_back(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            x = 9
            def foo(y):
                return y + 3
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            x = 200
            def foo(y):
                return y + 5
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.foo(0) == 5

    def test_deduperreloader_failure(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            x = 9
            def foo(y):
                return y + 3
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            broken broken broken
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.foo(0) == 3

    def test_deduperreloader_imported_mod(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            from os import environ
            def foo(n):
                pass
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            from os import environ
            def foo():
                environ._data
                return 1
        """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.foo() == 1


class AutoreloadClassMethodsDetectionSuite(unittest.TestCase):
    """
    Unit tests for autoreload testing logic
    """

    def test_autoreload_method(self):
        code1 = squish_text(
            """
            class C(n):
                def foo():
                    pass
        """
        )
        code2 = squish_text(
            """
            class C(n):
                def foo():
                    x = 1
                    return x
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)
        assert deduperreloader._to_autoreload.defs_to_reload == []
        assert "C" in deduperreloader._to_autoreload.children
        assert ["foo"] == list(
            d[0][0] for d in deduperreloader._to_autoreload.children["C"].defs_to_reload
        )
        assert deduperreloader._to_autoreload.children["C"].children == {}

    def test_autoreload_no_changes(self):
        code1 = squish_text(
            """
            class D(n):
                def foo():
                    pass
            def foo():
                pass
        """
        )
        code2 = squish_text(
            """
            class D(n):
                def foo():
                    pass
            def foo():
                pass
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)
        assert deduperreloader._to_autoreload.defs_to_reload == []
        assert deduperreloader._to_autoreload.children == {}

    def test_autoreload_add_function(self):
        code1 = squish_text(
            """
            class C:
                def foo():
                    pass
            def foo():
                pass
        """
        )
        code2 = squish_text(
            """
            class C:
                def foo():
                    x = 1
                    return x
                def bar():
                    return -1
            def foo():
                pass
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)
        assert deduperreloader._to_autoreload.defs_to_reload == []
        assert "C" in deduperreloader._to_autoreload.children
        assert ["foo", "bar"] == list(
            d[0][0] for d in deduperreloader._to_autoreload.children["C"].defs_to_reload
        )
        assert len(deduperreloader._to_autoreload.children["C"].children) == 0

    def test_autoreload_remove_method(self):
        code1 = squish_text(
            """
            class C:
                def foo():
                    pass
            def foo():
                pass
        """
        )
        code2 = squish_text(
            """
            class C:
                pass
            def foo():
                pass
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)

    def test_autoreload_remove_class(self):
        code1 = squish_text(
            """
            class C:
                def foo():
                    pass
            def foo():
                pass
        """
        )
        code2 = squish_text(
            """
            def foo():
                pass
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)

    def test_autoreload_add_class_in_class(self):
        code1 = squish_text(
            """
            class C:
                pass
            def foo():
                pass
        """
        )
        code2 = squish_text(
            """
            class C:
                class D:
                    pass
            def foo():
                pass
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)

    def test_autoreload_add_method_in_class_in_class(self):
        code1 = squish_text(
            """
            class C:
                class D:
                    x = 1
            def foo():
                pass
        """
        )
        code2 = squish_text(
            """
            class C:
                class D:
                    x = 1
                    def foo():
                        return 1
            def foo():
                pass
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)
        assert deduperreloader._to_autoreload.defs_to_reload == []
        assert list(deduperreloader._to_autoreload.children.keys()) == ["C"]
        assert deduperreloader._to_autoreload.children["C"].defs_to_reload == []
        assert list(deduperreloader._to_autoreload.children["C"].children.keys()) == [
            "D"
        ]
        assert list(
            d[0][0]
            for d in deduperreloader._to_autoreload.children["C"]
            .children["D"]
            .defs_to_reload
        ) == ["foo"]
        assert deduperreloader._to_autoreload.children["C"].children["D"].children == {}

    def test_autoreload_add_var_and_method_in_class_in_class(self):
        code1 = squish_text(
            """
            class C:
                class D:
                    pass
            def foo():
                pass
        """
        )
        code2 = squish_text(
            """
            class C:
                class D:
                    x = 1
                    def foo():
                        return 1
            def foo():
                pass
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)

    def test_autoreload_add_method_in_class_in_class_pass(self):
        code1 = squish_text(
            """
            class C:
                class D:
                    pass
            def foo():
                pass
        """
        )
        code2 = squish_text(
            """
            class C:
                class D:
                    def foo():
                        return 1
            def foo():
                pass
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)
        assert deduperreloader._to_autoreload.defs_to_reload == []
        assert list(deduperreloader._to_autoreload.children.keys()) == ["C"]
        assert deduperreloader._to_autoreload.children["C"].defs_to_reload == []
        assert list(deduperreloader._to_autoreload.children["C"].children.keys()) == [
            "D"
        ]
        assert list(
            d[0][0]
            for d in deduperreloader._to_autoreload.children["C"]
            .children["D"]
            .defs_to_reload
        ) == ["foo"]
        assert deduperreloader._to_autoreload.children["C"].children["D"].children == {}

    def test_autoreload_add_method_in_class_in_class_more(self):
        code1 = squish_text(
            """
            class C:
                class D:
                    pass
            def foo():
                pass
        """
        )
        code2 = squish_text(
            """
            class C:
                def bar():
                    return -1
                class D:
                    def foo():
                        return 1
            def foo():
                pass
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)
        assert deduperreloader._to_autoreload.defs_to_reload == []
        assert list(deduperreloader._to_autoreload.children.keys()) == ["C"]
        assert list(
            d[0][0] for d in deduperreloader._to_autoreload.children["C"].defs_to_reload
        ) == ["bar"]
        assert list(deduperreloader._to_autoreload.children["C"].children.keys()) == [
            "D"
        ]
        assert list(
            d[0][0]
            for d in deduperreloader._to_autoreload.children["C"]
            .children["D"]
            .defs_to_reload
        ) == ["foo"]
        assert deduperreloader._to_autoreload.children["C"].children["D"].children == {}

    def test_autoreload_add_method_in_class_in_class_with_members(self):
        code1 = squish_text(
            """
            class C:
                class D:
                    pass
            def foo():
                pass
        """
        )
        code2 = squish_text(
            """
            class C:
                def bar():
                    return -1
                class D:
                    s = 2
                    def foo():
                        return 1
            def foo():
                pass
        """
        )
        deduperreloader = DeduperTestReloader()
        ast_1 = ast.parse(code1)
        ast_2 = ast.parse(code2)

        assert deduperreloader.detect_autoreload(ast_1, ast_2)


class AutoreloadReliabilitySuite(ShellFixture):
    def test_autoreload_class_basic(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            x = 9
            class C:
                def foo():
                    return 1
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            x = 9
            class C:
                def foo():
                    return 1
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.C.foo() == 1

    def test_remove_overridden_method(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            class A:
                def foo(self):
                    return 1
            class B(A):
                def foo(self):
                    return 42
        """
        )
        self.shell.run_code(f"from {mod_name} import B; b = B()")
        self.shell.run_code("assert b.foo() == 42")
        self.write_file(
            mod_fn,
            """
            class A:
                def foo(self):
                    return 1
            class B(A):
                pass
            """,
        )
        self.shell.run_code("assert b.foo() == 1")

    def test_autoreload_class_use_outside_func(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            x = 9
            class C:
                def foo():
                    return 1
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            x = 9
            class C:
                def foo():
                    return 1+x
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.C.foo() == 10

    def test_autoreload_class_use_class_member(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            x = 9
            class C:
                def foo():
                    return 1
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            class C:
                x = 9
                def foo():
                    return 1+C.x
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.C.foo() == 10

    def test_autoreload_class_pass(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            x = 9
            class C:
                pass
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            class C:
                x = 9
                def foo():
                    return 1+C.x
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.C.foo() == 10

    def test_autoreload_class_ellipsis(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            x = 9
            class C:
                ...
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            class C:
                x = 9
                def foo():
                    return 1+C.x
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.C.foo() == 10

    def test_autoreload_class_default_autoreload(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            class C:
                x = 9
                def foo():
                    return 1+C.x
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            class C:
                x = 20
                def foo():
                    return 1+C.x
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.C.foo() == 21

    def test_autoreload_class_nested(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            class C:
                x = 9
                class D:
                    def foo():
                        pass
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            class C:
                x = 9
                class D:
                    def foo():
                        return 10
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.C.D.foo() == 10

    def test_autoreload_class_nested2(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            class C:
                x = 9
                def c():
                    return 1
                class D:
                    def foo():
                        pass
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            class C:
                x = 9
                def c():
                    return 1
                class D:
                    def foo():
                        return 10
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.C.D.foo() == 10
        assert mod.C.c() == 1

    def test_autoreload_class_nested3(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            class C:
                x = 9
                def c():
                    return 1
                class D:
                    def foo():
                        pass
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            class C:
                x = 9
                def c():
                    return 13
                class D:
                    def foo():
                        return 10
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.C.D.foo() == 10
        assert mod.C.c() == 13

    def test_autoreload_new_class_added(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            class C:
                x = 9
                def c():
                    return 1
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            class C:
                x = 9
                def c():
                    return 13
            class D:
                def c():
                    return 1
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.C.c() == 13
        assert mod.D.c() == 1

    def test_autoreload_class_nested_default(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            class C:
                pass
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            class C:
                x = 9
                def c():
                    return 13
                class D:
                    def foo():
                        return 10
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.C.D.foo() == 10
        assert mod.C.c() == 13

    def test_autoreload_class_nested_using_members(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            class C:
                x = 9
                def c():
                    return 13
                class D:
                    def foo():
                        return 10
            """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            class C:
                x = 9
                def c():
                    return 13
                class D:
                    def foo():
                        return 10 + C.x
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.C.D.foo() == 19
        assert mod.C.c() == 13

    def test_autoreload_class_nested_using_members_ellipsis(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            class C:
                x = 9
                def c():
                    return 13
                class D:
                    def foo():
                        ...
            """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            class C:
                x = 9
                def c():
                    return 13
                class D:
                    def foo():
                        return 10 + C.x
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.C.D.foo() == 19
        assert mod.C.c() == 13

    def test_method_decorators_no_changes(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_file = self.new_module(
            """
            class Foo:
                @classmethod
                def bar(cls):
                    return 0
                    
                @classmethod
                def foo(cls):
                    return 42 + cls.bar()
            
            foo = Foo.foo
            """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.foo() == 42

        self.shell.run_code(f"assert {mod_name}.foo() == 42")
        self.write_file(
            mod_file,
            """
            class Foo:
                @classmethod
                def bar(cls):
                    return 0
            
                @classmethod
                def foo(cls):
                    return 42 + cls.bar()
            
            foo = Foo.foo
            """,
        )
        self.shell.run_code(f"assert {mod_name}.foo() == 42")

    def test_method_decorators_no_changes1(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_file = self.new_module(
            """
            class Foo:
                @classmethod
                def bar(cls):
                    return 0
                    
                @classmethod
                def foo(cls):
                    return 42 + cls.bar()
            
            foo = Foo.foo
            """
        )

        self.shell.run_code(f"from {mod_name} import foo")
        self.shell.run_code("assert foo() == 42")
        self.write_file(
            mod_file,
            """
            class Foo:
                @classmethod
                def bar(cls):
                    return 0
            
                @classmethod
                def foo(cls):
                    return 42 + cls.bar()
            
            foo = Foo.foo
            """,
        )
        self.shell.run_code("assert foo() == 42")

    def test_method_classmethod_one_change(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_file = self.new_module(
            """
            class Foo:
                @classmethod
                def bar(cls):
                    return 0
                    
                @classmethod
                def func(cls):
                    return 42 + cls.bar()
            
            func = Foo.func
            """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code(f"assert {mod_name}.func() == 42")
        self.write_file(
            mod_file,
            """
            class Foo:
                @classmethod
                def bar(cls):
                    return 1
            
                @classmethod
                def func(cls):
                    return 42 + cls.bar()
            
            func = Foo.func
            """,
        )
        mod = sys.modules[mod_name]
        self.shell.run_code(f"assert {mod_name}.func() == 43")

    def test_method_staticmethod_one_change(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_file = self.new_module(
            """
            class Foo:
                @staticmethod
                def bar():
                    return 0
                    
                @staticmethod
                def func():
                    return 42 + Foo.bar()
            
            func = Foo.func
            """
        )
        self.shell.run_code(f"from {mod_name} import func")
        self.shell.run_code("assert func() == 42")
        self.write_file(
            mod_file,
            """
            class Foo:
                @staticmethod
                def bar():
                    return 1
            
                @staticmethod
                def func():
                    return 42 + Foo.bar()
            
            func = Foo.func
            """,
        )
        self.shell.run_code("assert func() == 43")

    def test_autoreload_class_default_args(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            x = 42
            class Foo:
                def foo(self, y): return y
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            x = 42
            class Foo:
                def foo(self, y=x): return y
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        obj = mod.Foo()
        assert obj.foo(2) == 2
        assert obj.foo() == 42

    def test_autoreload_class_change_default_args(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            x = 42
            class Foo:
                def foo(y): return y
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        self.write_file(
            mod_fn,
            """
            x = 44
            class Foo:
                def foo(y=x): return y
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        mod = sys.modules[mod_name]
        assert mod.Foo.foo() == 44

    def test_autoreload_class_new_class(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            x = 42
            class Foo:
                def foo(y=x): return y
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        prev_foo = self.shell.user_ns[mod_name].Foo.foo
        self.write_file(
            mod_fn,
            """
            x = 42
            class Foo:
                def foo(y=x): return y
            class C:
                def foo(): return 200
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        self.assertIs(prev_foo, self.shell.user_ns[mod_name].Foo.foo)
        mod = sys.modules[mod_name]
        assert mod.Foo.foo() == 42
        assert mod.C.foo() == 200

    def test_autoreload_overloaded_vars(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            x = 42
            class Foo:
                pass
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        mod = sys.modules[mod_name]
        self.write_file(
            mod_fn,
            """
            x = 44
            class Foo:
                def foo(): 
                    x = 2
                    return x
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        assert mod.Foo.foo() == 2

    def test_autoreload_overloaded_vars2(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            x = 42
            def foo():
                return x
        """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        mod = sys.modules[mod_name]
        self.write_file(
            mod_fn,
            """
            x = 44
            def foo(): 
                x = 2
                return x
            """,
        )
        self.shell.run_code("pass")
        self.assertIn(mod_name, self.shell.user_ns)
        assert mod.foo() == 2


class DecoratorPatchingSuite(ShellFixture):
    """
    Unit tests for autoreload patching logic
    """

    def test_modify_property(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_file = self.new_module(
            """
            class Foo:
                @property
                def foo(self):
                    return 42
            """
        )
        self.shell.run_code(f"from {mod_name} import Foo")
        self.shell.run_code("foo = Foo()")
        self.shell.run_code("assert foo.foo == 42")
        self.write_file(
            mod_file,
            """
            class Foo:
                @property
                def foo(self):
                    return 43
            """,
        )
        self.shell.run_code("pass")
        self.shell.run_code("assert foo.foo == 43")

    def test_method_decorator(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_file = self.new_module(
            """
            def incremented(f):
                return lambda *args: f(*args) + 1

            class Foo:
                @classmethod
                @incremented
                def foo(cls):
                    return 42
            
            foo = Foo.foo
            """
        )
        self.shell.run_code(f"from {mod_name} import foo")
        self.shell.run_code("assert foo() == 43")
        self.write_file(
            mod_file,
            """
            def incremented(f):
                return lambda *args: f(*args) + 1

            class Foo:
                @classmethod
                def foo(cls):
                    return 42
            
            foo = Foo.foo
            """,
        )
        self.shell.run_code("assert foo() == 42")

    def test_method_modified_decorator(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_file = self.new_module(
            """
            def incremented(f):
                return lambda *args: f(*args) + 1

            class Foo:
                @classmethod
                @incremented
                def foo(cls):
                    return 42
            
            foo = Foo.foo
            """
        )
        self.shell.run_code(f"from {mod_name} import foo")
        self.shell.run_code("assert foo() == 43")
        self.write_file(
            mod_file,
            """
            def incremented(f):
                return lambda *args: f(*args) + 0

            class Foo:
                @classmethod
                @incremented
                def foo(cls):
                    return 42
            
            foo = Foo.foo
            """,
        )
        self.shell.run_code("assert foo() == 42")

    def test_function_decorators(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_file = self.new_module(
            """
            def incremented(f):
                return lambda *args: f(*args) + 1
            
            @incremented
            def foo():
                return 42
            """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        mod = sys.modules[mod_name]
        assert mod.foo() == 43
        self.write_file(
            mod_file,
            """
            def incremented(f):
                return lambda *args: f(*args) + 1

            def foo():
                return 42
            """,
        )
        self.shell.run_code("pass")
        assert mod.foo() == 42
        self.write_file(
            mod_file,
            """
            def incremented(v):
                def deco(f):
                    return lambda *args: f(*args) + v
                return deco

            @incremented(2)
            def foo():
                return 43
            """,
        )
        self.shell.run_code("pass")
        assert mod.foo() == 45

    def test_method_decorators_again(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_file = self.new_module(
            """
            class Foo:
                @classmethod
                def bar(cls):
                    return 0
                    
                @classmethod
                def foo(cls):
                    return 42 + cls.bar()
            
            foo = Foo.foo
            """
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        mod = sys.modules[mod_name]
        assert mod.foo() == 42
        self.write_file(
            mod_file,
            """
            class Foo:
                @classmethod
                def bar(cls):
                    return 1
            
                @classmethod
                def foo(cls):
                    return 42 + cls.bar()
            
            foo = Foo.foo
            """,
        )
        self.shell.run_code("pass")
        assert mod.Foo.foo() == 43
        assert mod.foo() == 43

    # This test verifies that the correct globals are used when patching a function
    # decorated by a function from another module. For this example, we should always
    # use <mod_name>.__dict__ as the global environment when patching foo, which comes
    # from <mod_name>, and never <other_mod_name>.__dict__, which is what we would get
    # if we use foo.__globals__ after it has been decorated.
    def test_function_decorator_from_other_module(self):
        self.shell.magic_autoreload("2")
        other_mod_name, _ = self.new_module(
            """
            import functools

            def incremented(f):
                @functools.wraps(f)
                def wrapper(*args, **kwargs):
                    return f(*args, **kwargs) + 1
                return wrapper
            """
        )
        mod_name, mod_file = self.new_module(
            f"""
            from {other_mod_name} import incremented as deco
            @deco
            def foo():
                return 42
            """
        )
        self.shell.run_code(f"from {mod_name} import foo")
        self.shell.run_code("assert foo() == 43")
        self.write_file(
            mod_file,
            f"""
            from {other_mod_name} import incremented as deco
            @deco
            def foo():
                return 43
            """,
        )
        self.shell.run_code("assert foo() == 44")

    def test_decorators_with_freevars(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_file = self.new_module(
            """
            def decorate(func):
                free_var_x = "x"
                free_var_y = "y"
                def wrapper(*args, **kwargs):
                    return func(*args, **kwargs), free_var_x, free_var_y
                return wrapper
            @decorate
            def f():
                return "v1"
            """
        )
        self.shell.run_code(f"from {mod_name} import f")
        mod = sys.modules[mod_name]
        assert mod.f() == ("v1", "x", "y")
        self.write_file(
            mod_file,
            """
            def decorate(func):
                free_var_ax = "ax"
                free_var_by = "by"
                def wrapper(*args, **kwargs):
                    return func(*args, **kwargs), free_var_ax, free_var_by
                return wrapper
            @decorate
            def f():
                return "v2"
            """,
        )
        self.shell.run_code("pass")
        val = mod.f()
        assert val == ("v2", "ax", "by"), val


class TestAutoreloadEnum(ShellFixture):
    def test_reload_enums(self):
        self.shell.magic_autoreload("2")
        mod_name, mod_fn = self.new_module(
            """
            from enum import Enum
            class MyEnum(Enum):
                A = 'A'
                B = 'B'
            """,
        )
        self.shell.run_code("import %s" % mod_name)
        self.shell.run_code("pass")
        mod = sys.modules[mod_name]
        self.write_file(
            mod_fn,
            """
            from enum import Enum
            class MyEnum(Enum):
                A = 'A'
                B = 'B'
                C = 'C'
            """,
        )
        self.shell.run_code("pass")
        assert mod.MyEnum.C.value == "C"


if __name__ == "__main__":
    unittest.main()



from IPython.core.tips import _tips

import os
import pytest

all_tips = _tips["random"] + list(_tips["every_year"].values())


@pytest.mark.skipif(os.name != "nt", reason="Windows console may crash with Unicode")
@pytest.mark.parametrize("tip", all_tips)
def test_tips(tip):
    assert tip.isascii()
